# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Mapping backend based on an in-memory geo-referenced numpy array."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any

import numpy as np
import numpy.typing as npt
from affine import Affine
from pyproj import CRS

from weitsicht.exceptions import CRSInputError, CRSnoZaxisError
from weitsicht.geometry.interpolation_bilinear import bilinear_interpolation
from weitsicht.geometry.intersection_bilinear import multilinear_poly_intersection
from weitsicht.geometry.line_grid_intersection import (
    line_grid_intersection_points,
    raster_index_p1_p2,
    vector_projection,
)
from weitsicht.mapping.base_class import MappingBase, MappingType
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import (
    ArrayN_,
    ArrayNx2,
    ArrayNx3,
    ArrayNxN,
    Issue,
    MappingResult,
    MappingResultSuccess,
    MaskN_,
    ResultFailure,
    to_array_nx3,
)

__all__ = ["MappingGeorefArray"]

logger = logging.getLogger(__name__)

neighbour_offsets = [
    np.array([0, 0]),
    np.array([1, -1]),
    np.array([1, 0]),
    np.array([1, 1]),
    np.array([0, 1]),
    np.array([0, -1]),
    np.array([-1, -1]),
    np.array([-1, 0]),
    np.array([-1, 1]),
]


# def tracer_bilinear_patch(points, pos_ray, ray):
#     bilinear_patch = BilinearPatch(*points)
#
#     ray_vec = Vector.from_vec(ray)
#     ray_vec.normalize()
# cover
#     pos = Vector.from_vec(pos_ray)
#     uv = bilinear_patch.ray_patch_intersection(pos_ray=pos,
#                                                ray=ray_vec)
#
#     if uv is not None:
#         return bilinear_patch.srf_eval(uv.x, uv.y).vec
#
#     return None


class MappingGeorefArray(MappingBase):
    """Mapping backend using an in-memory geo-referenced raster array.

    This backend is typically used internally (e.g. for raster window caching) and supports
    height sampling and ray intersections using bilinear patches.
    """

    def __init__(self, raster_array: ArrayNxN, geo_transform: Affine, crs: CRS | None):
        """Create a mapper backed by a numpy array and an affine geo-transform.

        :param raster_array: Raster data array (rows × cols) holding height values.
        :type raster_array: ArrayNxN
        :param geo_transform: Affine transformation mapping pixel indices to raster CRS coordinates.
        :type geo_transform: Affine
        :param crs: CRS of the raster array, defaults to ``None``.
        :type crs: CRS | None
        :raises CRSnoZaxisError: If a specified CRS does not define a Z axis.
        """
        super().__init__()

        self._raster_array = raster_array
        self._raster_array.setflags(write=False)
        self._transform: Affine = geo_transform
        self.area_flag = True

        # User CRS can override CRS from dataset
        self._crs = crs

        if self._crs is not None:
            if len(self._crs.axis_info) < 3:
                logger.error("CRS has no Z axis defined")
                raise CRSnoZaxisError("CRS has no Z axis defined")

        self._width = self._raster_array.shape[1]
        self._height = self._raster_array.shape[0]

    @property
    def transform(self):
        """Return the affine geo-transform.

        :return: Affine geo-transform.
        :rtype: Affine
        """
        return self._transform

    @property
    def width(self):
        """Return raster width in pixels."""
        return self._width

    @property
    def height(self):
        """Return raster height in pixels."""
        return self._height

    @classmethod
    def from_dict(cls, mapper_dict: Mapping[str, Any]) -> MappingGeorefArray:
        """Create a mapper from a configuration dictionary.

        This is currently not supported for :class:`MappingGeorefArray`.

        :raises NotImplementedError: Always raised.
        """
        raise NotImplementedError("This is not implemented for GeorefArray")

    @property
    def type(self) -> MappingType:
        """Return the mapper type."""
        return MappingType.GeoreferencedNumpyArray

    @property
    def param_dict(self) -> dict:
        """Return the mapper configuration dictionary.

        This is currently not supported for :class:`MappingGeorefArray`.

        :raises NotImplementedError: Always raised.
        """
        raise NotImplementedError("This is not implemented for GeorefArray")

    def pixel_to_coordinate(self, px_row: ArrayN_, px_col: ArrayN_) -> tuple:
        """Convert pixel indices to coordinates in raster CRS.

        :param px_row: Pixel row coordinate
        :type px_row: ``ArrayN_``
        :param px_col: Pixel column coordinate
        :type px_col: ``ArrayN_``
        :return: ``(x, y)`` coordinates in raster CRS.
        :rtype: tuple[``ArrayN_``, ``ArrayN_``]
        """
        return self.transform * np.vstack((px_col, px_row))

    def coordinate_to_pixel(self, x: ArrayN_, y: ArrayN_) -> tuple[ArrayN_, ArrayN_]:
        """Convert raster CRS coordinates to pixel indices.

        :param x: X coordinate in raster CRS.
        :type x: ``ArrayN_``
        :param y: Y coordinate in raster CRS.
        :type y: ``ArrayN_``
        :return: ``(row, col)`` pixel coordinates.
        :rtype: tuple[``ArrayN_``, ``ArrayN_``]
        """
        pix_col, pix_row = ~self.transform * np.vstack((x, y))
        return np.asarray(pix_row), np.asarray(pix_col)

    def pixel_valid(self, px_row: ArrayN_, px_col: ArrayN_) -> MaskN_:
        """Vectorized validity check for pixel coordinates.

        :param px_row: Pixel row coordinate
        :type px_row: ``ArrayN_``
        :param px_col: Pixel column coordinate
        :type px_col: ``ArrayN_``
        :return: Boolean mask for valid pixels.
        :rtype: ``MaskN_``
        """
        # 0    1    2    3    4
        # |----|----|----|----|
        #                   x
        # self.height = 100
        # px_row index can be from 0 to 99.9999

        # 0 <= px_row < self.height and 0 <= px_col < self.width
        # return bool(np.all((0 <= px_row, px_row < self.height, 0 <= px_col, px_col < self.width)))
        valid: MaskN_ = (px_row >= 0) & (px_row < self.height) & (px_col >= 0) & (px_col < self.width)
        return valid

    def pixel_valid_index_ray_bilinear_shifted(self, px_row: ArrayN_, px_col: ArrayN_) -> np.ndarray:
        """Return indices of pixels valid for bilinear ray intersection.

        This check assumes pixel coordinates are already shifted by ``0.5``. Border pixels are excluded
        because bilinear patches require 4 surrounding corner points.

        :param px_row: Pixel row coordinate already shifted by 0.5
        :type px_row: ``ArrayN_``
        :param px_col: Pixel column coordinate already shifted by 0.5
        :type px_col: ``ArrayN_``
        :return: Indices of valid pixels.
        :rtype: numpy.ndarray
        """
        # 0    1    2    3    4
        # |----|----|----|----|
        #                    x
        # self.height = 100
        # px_row index can be from 0.5 to 99.5
        # for the bilinear ray intersection the border region
        # is not valid as there the points are not within 4 points

        # here we say smaller than that because otherwise if a ray is vertical
        # and on the exact upper boundary there is a problem
        # bool_row = (px_row < self.height - 1) & (px_row >= 0)
        # bool_col = (px_col < self.width - 1) & (px_col >= 0)
        # index_combined = np.logical_and(bool_row, bool_col)
        return np.flatnonzero((px_row < self.height - 1) & (px_row >= 0) & (px_col < self.width - 1) & (px_col >= 0))

    def coordinate_on_raster(self, x_crs: ArrayN_, y_crs: ArrayN_) -> bool:
        """Test if given coordinates in raster CRS are on the raster

        :param x_crs: X coordinate
        :type x_crs: ``ArrayN_``
        :param y_crs: Y coordinate
        :type y_crs: ``ArrayN_``
        :return: ``True`` if all coordinates are within raster bounds.
        :rtype: bool
        """

        pix_row, pix_col = self.coordinate_to_pixel(x_crs, y_crs)
        return bool(np.all(self.pixel_valid(px_row=pix_row, px_col=pix_col)))

    def map_coordinates_from_rays_old_sampling(
        self, ray_vectors_crs_s: ArrayNx3, ray_start_crs_s: ArrayNx3, crs_s: CRS
    ) -> ArrayNx3 | None:  # pragma: no cover
        """Legacy ray intersection method using dense sampling (deprecated).

        :param ray_vectors_crs_s: Ray direction vectors (N×3).
        :type ray_vectors_crs_s: ArrayNx3
        :param ray_start_crs_s: Ray start points (N×3).
        :type ray_start_crs_s: ArrayNx3
        :param crs_s: CRS of the input rays.
        :type crs_s: CRS
        :return: Intersection coordinates or ``None`` if no intersection was found.
        :rtype: ArrayNx3 | None
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        # TODO The length and start of the sampling could be smart estimated, as well the sampling resolution
        # TODO we could as well make the length of the sampling vector in a for loop, so to extend distance

        coo_trafo = CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)

        intersection_point = []
        for idx_ray, ray in enumerate(ray_vectors_crs_s):
            # 1. Sample the ray vectors to form an array of points.
            # The sampling is not really of importance here as here only a
            # array is loaded which does not know how big the
            # source raster could be. And its only a numpy indexing. So we use 1cm.

            # TODO calc max distance possible for ray
            # max distance use minimum of height of georef array
            # intersection_plane(line_vec: np.ndarray, line_point: np.array([0,0,self._raster_array.min()]),
            # plane_point: np.ndarray, plane_normal: np.ndarray | None = None)

            dist_multi = np.arange(0, 900, 0.1)
            # This forms an array filled with unit vectors in the direction of the ray
            # which is than scaled by the sampled distance to get points along the ray
            ray_points_crs_s = (np.zeros((dist_multi.shape[0], 3)) + ray / np.linalg.norm(ray)).T * dist_multi
            # Now we translate the points to the correct location in 3d space
            ray_points_crs_s = ray_points_crs_s.T + ray_start_crs_s[idx_ray, :]

            # 2. Transform that points into the raster crs and check if all points are inside the raster
            if coo_trafo is not None:
                ray_points_crs_array = coo_trafo.transform(ray_points_crs_s)
            else:
                ray_points_crs_array = ray_points_crs_s * 1.0

            # Transform position of ray points in array crs to array col and row by the affine transformation
            # col, row = ~self._transform * ray_points_crs_array[:, :2].T
            row, col = self.coordinate_to_pixel(ray_points_crs_array[:, 0], ray_points_crs_array[:, 1])

            # 3. Find the first approx intersection of the ray and the raster

            # Only cells, where one of the corner points is higher and one lower as the sample point,
            # are potential intersection points
            # If all corner points are lower there can be no intersection in a bilinear patch

            # TODO check if outside raster? Or at least check if an intersection was found inside the raster
            # center_cells = np.hstack((np.floor(_row)+0.5, np.floor(_col)+0.5))
            # Find unique cells which can be tested. As sampling can lead to give the same cell more times
            # unique_rows = np.unique(center_cells, axis=0)

            # TODO: What would happen if one ray point is exactly on a raster edge or line?

            # TODO: A Problem is when using the same CRS and the ray is exactly on the line
            # Then for example if line vec = [0,0.3,-1] the col lower and col upper are equal
            # We should always get all surroundings cells not only via ceil and floor
            row_lower = np.floor(row).astype(int)
            row_upper = np.ceil(row).astype(int)
            col_lower = np.floor(col).astype(int)
            col_upper = np.ceil(col).astype(int)
            r1 = self._raster_array[row_lower, col_lower]
            r2 = self._raster_array[row_lower, col_upper]
            r3 = self._raster_array[row_upper, col_upper]
            r4 = self._raster_array[row_upper, col_lower]

            r_corners = np.vstack((r1, r2, r3, r4))

            cells_min = np.min(r_corners, axis=0)
            cells_max = np.max(r_corners, axis=0)

            # To find the indices where a possible intersection can occur,
            # 1) we check if a point is within the min an max of a cell and take this one and one before
            # Because the intersection could already happen before that point

            index_to_search: list[int] = []

            index_between: npt.NDArray[np.int_] = np.flatnonzero(
                (cells_max >= ray_points_crs_array[:, 2]) & (ray_points_crs_array[:, 2] >= cells_min)
            )

            if index_between.size != 0:
                index_to_search += [int(x) for x in index_between]
                # Add also the index before to it
                index_to_search += [int(x) - 1 for x in index_between]

            # 2) we test always 2 point of the ray if any of the raster values are within that 2 points

            stack_2_points = np.vstack((ray_points_crs_array[0:-1, 2], ray_points_crs_array[1:, 2]))
            stack_2_points_max = np.max(stack_2_points, axis=0)
            stack_2_points_min = np.max(stack_2_points, axis=0)

            stack_2_cells_max = np.vstack((cells_max[0:-1], cells_max[1:]))
            stack_2_cells_max_lower = np.min(stack_2_cells_max, axis=0)

            stack_2_cells_min = np.vstack((cells_max[0:-1], cells_max[1:]))
            stack_2_cells_min_higher = np.max(stack_2_cells_min, axis=0)

            index_between: npt.NDArray[np.int_] = np.flatnonzero(
                (stack_2_points_max >= stack_2_cells_max_lower) & (stack_2_points_min <= stack_2_cells_min_higher)
            )

            if index_between.size != 0:
                index_to_search += [int(x) for x in index_between]
                # Add also the index after should be added as we here get the index of the first cell
                # If we say that sampling is at least smaller than the raster sampling this is fine
                index_to_search += [int(x) + 1 for x in index_between]

            # Get unique index of cells to test with bilinear patches
            index_to_search = sorted(list(set(index_to_search)))

            # Now test all index starting from the first one till we have found an intersection
            # 4) Refine the intersection to be precise

            if len(index_to_search) > 0:
                # TODO: we could filter as well with numpy operation using r1 to r4
                #  instead of checking which cells are already looked. Should anyhow make not so much speed difference
                # We will only test cells which have not been tested. As the sampling can deliver same cells
                cells_already_looked = []

                # We now need to transform the four points into a cartesian space to run the bilinear intersection
                # TODO Maybe its better to transform all points of the for loop at once for runtime?

                uv = None
                for _index in index_to_search:
                    if uv is not None:
                        break
                    cell_corners_initial = np.array(
                        [
                            [row_lower[_index], col_lower[_index]],
                            [row_upper[_index], col_lower[_index]],
                            [row_lower[_index], col_upper[_index]],
                            [row_upper[_index], col_upper[_index]],
                        ]
                    )

                    for neighbour_offset in neighbour_offsets:
                        # The order of the points which create the bilinear patch is very important!!!
                        # the 4th point should be the one on the opposite side of point 1
                        cell_corners = cell_corners_initial + neighbour_offset

                        if cell_corners.tolist() in cells_already_looked:
                            continue

                        cells_already_looked.append(cell_corners.tolist())

                        # Get for corner pixel the coordinates in raster crs
                        x, y = self.pixel_to_coordinate(px_row=cell_corners[:, 0], px_col=cell_corners[:, 1])

                        # Create 3D point from pixel coordinates and raster value (height)
                        cell_3d_points = np.vstack(
                            (
                                x,
                                y,
                                self._raster_array[cell_corners[:, 0], cell_corners[:, 1]],
                            )
                        ).T

                        # Transform back to source crs.
                        if coo_trafo is not None:
                            patch_points_crs_s = coo_trafo.transform(cell_3d_points, direction="inverse")
                        else:
                            patch_points_crs_s = cell_3d_points * 1.0

                        # TODO Should we check if points coplanar?
                        # I think we would not gain any speed.

                        intersect = multilinear_poly_intersection(
                            patch_points_crs_s, p=ray_start_crs_s[idx_ray, :], r=ray
                        )

                        if intersect is not None:
                            intersection_point.append(intersect)
                            break

        if len(intersection_point) > 0:
            return np.array(intersection_point)
        return None

    def map_coordinates_from_rays(
        self,
        ray_vectors_crs_s: ArrayNx3,
        ray_start_crs_s: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map coordinates by intersecting rays with the raster surface.

        :param ray_vectors_crs_s: Ray direction vectors (N×3).
        :type ray_vectors_crs_s: ArrayNx3
        :param ray_start_crs_s: Ray start points (N×3), same shape as ``ray_vectors_crs_s``.
        :type ray_start_crs_s: ArrayNx3
        :param crs_s: CRS of the input rays, defaults to ``None``.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing intersection coordinates and a validity mask.
        :rtype: MappingResult
        :raises ValueError: If input arrays have incompatible shapes.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        _ray_vectors_crs_s = to_array_nx3(ray_vectors_crs_s)
        _ray_start_crs_s = to_array_nx3(ray_start_crs_s)

        if _ray_start_crs_s.shape != _ray_vectors_crs_s.shape:
            raise ValueError("Start of rays and rays have to be same size")

        # TODO The length and start of the sampling could be smart estimated, as well the sampling resolution
        # TODO we could as well make the length of the sampling vector in a for loop, so to extend distance

        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        intersection_point = []
        valid_mask = []
        inside_mask = []
        for idx_ray, ray in enumerate(_ray_vectors_crs_s):
            # 1. Sample the ray vectors to get intersection lines of 50 meter.

            # TODO calc max distance possible for ray
            # max distance use minimum of height of georef array
            # intersection_plane(line_vec: np.ndarray, line_point: np.array([0,0,self._raster_array.min()]),

            # the maximum dist is for now 80km
            # Will walk in steps of 50 meter
            # we assume that within 50 meter any distortion by different projections will have not really an influence
            # will break iteration if the ray than is anyhow lower as we would then have already missed the intersection
            dist_multi = np.arange(0, 8000, 50)

            # This forms an array filled with unit vectors in the direction of the ray
            # which is than scaled by the sampled distance to get points along the ray
            ray_points_crs_s = (np.zeros((dist_multi.shape[0], 3)) + ray / np.linalg.norm(ray)).T * dist_multi

            del dist_multi
            # Now we translate the points to the correct location in 3d space
            ray_points_crs_s = ray_points_crs_s.T + _ray_start_crs_s[idx_ray, :]

            # 2. Transform that points into the raster crs
            # TODO what happens here if transformation grids are used. That would make the vector wrong
            # There can be height grids as well distortion grids on x, y -> This would be anyhow a big problem
            # Maybe we could only use height from starting points, if height grids are used this would work
            # Also if there would be rays with very long distance from start x,y to end x,y what happens with
            # mapping scales?
            if coo_trafo is not None:
                ray_points_crs_array = coo_trafo.transform(ray_points_crs_s)
            else:
                ray_points_crs_array = ray_points_crs_s * 1.0

            # Transform position of ray points in array crs to array col and row by the affine transformation
            pix_row, pix_col = self.coordinate_to_pixel(ray_points_crs_array[:, 0], ray_points_crs_array[:, 1])

            pix_row -= 0.5
            pix_col -= 0.5

            found_intersection_for_ray = False
            at_least_once_on_raster = False
            # Iterate now over the pieces of that ray
            for _idx in range(len(pix_row) - 1):
                # get the cells touching the line

                # if all raster z values of the cells are lower as the second point of the ray line
                # then go directly to the next line
                # TODO test both points. Like rays looking upward to mountain?

                x_1, x_2, y_1, y_2 = raster_index_p1_p2(
                    [pix_row[_idx], pix_col[_idx]],
                    [pix_row[_idx + 1], pix_col[_idx + 1]],
                )

                # TODO we should check if it was any time within the raster
                # otherwise we should return Issue.OUTSIDE_RASTER
                # The ray part is still outside the raster
                if (x_1 < 0 and x_2 < 0) or (x_1 > self.height - 1 and x_2 > self.height - 1):
                    continue

                if (y_1 < 0 and y_2 < 0) or (y_1 > self.width - 1 and y_2 > self.width - 1):
                    continue

                # As we here check the rectangle at which the projected part of the ray goes through for height
                # We need to make sare to be within the borders here as well..
                x_1 = 0 if x_1 < 0 else x_1
                y_1 = 0 if y_1 < 0 else y_1
                x_2 = 0 if x_2 < 0 else x_2
                y_2 = 0 if y_2 < 0 else y_2

                x_1 = self.height - 1 if x_1 > self.height - 1 else x_1
                y_1 = self.width - 1 if y_1 > self.width - 1 else y_1
                x_2 = self.height - 1 if x_2 > self.height - 1 else x_2
                y_2 = self.width - 1 if y_2 > self.width - 1 else y_2

                # Todo rays looking up?
                if not ((x_1 == x_2) or (y_1 == y_2)):
                    if np.max(self._raster_array[x_1 : x_2 + 1, y_1 : y_2 + 1]) < ray_points_crs_array[_idx + 1, 2]:
                        continue
                else:
                    if (x_1 == x_2) and (y_1 == y_2):
                        if self._raster_array[x_1, y_1] < ray_points_crs_array[_idx + 1, 2]:
                            continue
                    elif x_1 == x_2:
                        if np.max(self._raster_array[x_1, y_1 : y_2 + 1]) < ray_points_crs_array[_idx + 1, 2]:
                            continue
                    elif y_1 == y_2:
                        if np.max(self._raster_array[x_1 : x_2 + 1, y_1]) < ray_points_crs_array[_idx + 1, 2]:
                            continue

                cells = line_grid_intersection_points(
                    [pix_row[_idx], pix_col[_idx]],
                    [pix_row[_idx + 1], pix_col[_idx + 1]],
                )

                # In this case, differently to simple interpolation
                # by location we can not allow pixel positions within 0.5 of the border
                # In this area the interpolation is changed but this is not working for the rays

                # We filter only the valid ones
                index_valid = self.pixel_valid_index_ray_bilinear_shifted(cells[:, 0], cells[:, 1])

                if index_valid.size == 0:
                    continue

                # This is just a helper to check if for single rays it was at least once on the raster
                # So that we can return issue=OUTSIDE_RASTER
                at_least_once_on_raster = True

                cells = cells[index_valid, :]

                # Only cells, where corner points are higher and one lower as the sample point,
                # are potential intersection points
                # If all corner points are lower or higher there can be no intersection in a bilinear patch
                # if all corner points are higher we have already a problem because we missed the intersection

                # TODO check if outside raster? Or at least check if an intersection was found inside the raster

                cells_row1 = cells + np.array([1, 0])
                cells_col1 = cells + np.array([0, 1])
                cells_row1_col1 = cells + np.array([1, 1])

                r1 = self._raster_array[cells[:, 0], cells[:, 1]]
                r2 = self._raster_array[cells_row1[:, 0], cells_row1[:, 1]]
                r3 = self._raster_array[cells_col1[:, 0], cells_col1[:, 1]]
                r4 = self._raster_array[cells_row1_col1[:, 0], cells_row1_col1[:, 1]]
                r_corners = np.vstack((r1, r2, r3, r4))

                # if all raster z values of the cells are lower as the second point of the ray line
                # then go directly to the next line
                # this should actually already checked before
                if np.max(r_corners) < ray_points_crs_array[_idx + 1, 2]:
                    continue

                cells = np.hstack((cells, r1.reshape(-1, 1)))
                cells_row1 = np.hstack((cells_row1, r2.reshape(-1, 1)))
                cells_col1 = np.hstack((cells_col1, r3.reshape(-1, 1)))
                cells_row1_col1 = np.hstack((cells_row1_col1, r4.reshape(-1, 1)))

                cells_min = np.min(r_corners, axis=0)
                cells_max = np.max(r_corners, axis=0)

                # Another check to avoid checking of cells which can not be the solution
                cells_lower_than_ray = np.logical_and(
                    cells_min > ray_points_crs_array[_idx, 2],
                    cells_min > ray_points_crs_array[_idx + 1, 2],
                )

                cells_higher_than_ray = np.logical_and(
                    cells_max < ray_points_crs_array[_idx, 2],
                    cells_max < ray_points_crs_array[_idx + 1, 2],
                )

                cells_valid = np.flatnonzero(np.logical_not(np.logical_or(cells_lower_than_ray, cells_higher_than_ray)))

                if cells_valid.size == 0:
                    continue

                cells = cells[cells_valid, :]
                cells_row1 = cells_row1[cells_valid, :]
                cells_col1 = cells_col1[cells_valid, :]
                cells_row1_col1 = cells_row1_col1[cells_valid, :]

                r1 = None
                r2 = None
                r3 = None
                r4 = None

                # Now we will calculate all corner points height diff to the normal intersection point with the ray
                # Here we use the original pix_row and pix_col without the border changing
                p1 = np.array([pix_row[_idx], pix_col[_idx], ray_points_crs_array[_idx, 2]])
                p2 = np.array(
                    [
                        pix_row[_idx + 1],
                        pix_col[_idx + 1],
                        ray_points_crs_array[_idx + 1, 2],
                    ]
                )

                # intersection_plane_mat_operation(p1,p2-p1,cells)

                prj_1 = vector_projection(p1, p2, cells)
                prj_2 = vector_projection(p1, p2, cells_col1)
                prj_3 = vector_projection(p1, p2, cells_row1_col1)
                prj_4 = vector_projection(p1, p2, cells_row1)

                # This vertical stuff is not working for vertical lines
                # inside = np.logical_and(prj_1[:, 2] <= cells_max, prj_1[:, 2] >= cells_min)
                # inside = np.logical_or(inside, np.logical_and(prj_2[:, 2] <= cells_max, prj_2[:, 2] >= cells_min))
                # inside = np.logical_or(inside, np.logical_and(prj_3[:, 2] <= cells_max, prj_3[:, 2] >= cells_min))
                # inside = np.logical_or(inside, np.logical_and(prj_4[:, 2] <= cells_max, prj_4[:, 2] >= cells_min))

                inside = np.all(abs(prj_1[:, 0:2] - cells[:, 0:2]) <= 2, axis=1)
                inside = np.logical_and(inside, np.all(abs(prj_2[:, 0:2] - cells_col1[:, 0:2]) <= 2, axis=1))
                inside = np.logical_and(
                    inside,
                    np.all(abs(prj_3[:, 0:2] - cells_row1_col1[:, 0:2]) <= 2, axis=1),
                )
                inside = np.logical_and(inside, np.all(abs(prj_4[:, 0:2] - cells_row1[:, 0:2]) <= 2, axis=1))

                index_to_check = np.flatnonzero(inside)
                # Iteration over all the cell which the current piece of the ray is in 2d touched
                if len(index_to_check) > 0:
                    for _index in index_to_check:
                        # We run the bilinear intersection directly in pixel coordinates
                        # TODO Should we check if points coplanar?

                        # The order of the points which create the bilinear patch is very important!!!
                        # the 4th point should be the one on the opposite side of point 1
                        patch_points_crs_m = np.array(
                            [
                                cells[_index, :],
                                cells_col1[_index, :],
                                cells_row1[_index, :],
                                cells_row1_col1[_index, :],
                            ]
                        )

                        intersect = multilinear_poly_intersection(
                            patch_points_crs_m,
                            p=p1,
                            r=(p2 - p1) / np.linalg.norm(p2 - p1),
                        )

                        if intersect is not None:
                            intersect[0] += 0.5
                            intersect[1] += 0.5
                            intersection_point.append(intersect)
                            found_intersection_for_ray = True
                            break

                    if found_intersection_for_ray:
                        break

            if not found_intersection_for_ray:
                intersection_point.append(np.array([0, 0, 0]))

            valid_mask.append(found_intersection_for_ray)
            inside_mask.append(at_least_once_on_raster)

        intersection_point_array = np.array(intersection_point)
        if intersection_point_array.shape != ray_vectors_crs_s.shape:
            #    # Should be already handled in few lines above
            raise RuntimeError("*Intersection points are not the same size as ray vectors. This should not happen")

        if not any(inside_mask):
            return ResultFailure(
                ok=False,
                error="All was outside the raster",
                issues={Issue.OUTSIDE_RASTER},
            )

        mask = np.array(valid_mask, dtype=bool)
        mask_index = np.flatnonzero(mask)
        if not np.any(mask):
            return ResultFailure(
                ok=False,
                error="No intersections where found",
                issues={Issue.NO_INTERSECTION},
            )

        # We now transform everything back
        # first get world coordinates from pixel coordinates in mapper crs
        x, y = self.pixel_to_coordinate(
            px_row=intersection_point_array[mask_index, 0],
            px_col=intersection_point_array[mask_index, 1],
        )
        intersection_point_array[mask_index, 0] = x
        intersection_point_array[mask_index, 1] = y

        interp_source_crs = np.empty(intersection_point_array.shape, dtype=float)
        interp_source_crs.fill(np.nan)

        # Transform back to origin crs.
        if coo_trafo is not None:
            interp_source_crs[mask_index, :] = coo_trafo.transform(
                intersection_point_array[mask_index, :], direction="inverse"
            )
        else:
            interp_source_crs[mask_index, :] = np.array(intersection_point_array[mask_index, :]) * 1.0

        issue = set()
        if not np.all(mask):
            issue.add(Issue.NO_INTERSECTION)
        if not all(inside_mask):
            issue.add(Issue.OUTSIDE_RASTER)
        return MappingResultSuccess(ok=True, coordinates=interp_source_crs, mask=mask, issues=issue)

    def map_heights_from_coordinates(
        self,
        coordinates_crs_s: ArrayNx3 | ArrayNx2,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Sample heights for given coordinates using bilinear interpolation.

        :param coordinates_crs_s: Coordinates to sample (N×2 or N×3).
        :type coordinates_crs_s: ArrayNx3 | ArrayNx2
        :param crs_s: CRS of the input coordinates, defaults to ``None``.
        :type crs_s: CRS | None
        :param transformer: Coordinate transformer to mapper CRS, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result containing sampled coordinates and a validity mask.
        :rtype: MappingResult
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        # Pure Coordinate Transformation. No Transformation of directions involved
        # 1) Transform coordinates if needed
        # 2) get index from transformed coordinates
        # 3) check if all indices are valid within the array limits
        # 4) get array values for valid index
        # 5) bilinear interpolation in pixel crs
        # 6) Assign estimated height to mapper crs points
        # 6) transform 3d points back

        _coordinates_crs_s = to_array_nx3(coordinates_crs_s, 0.0)

        # Transformer
        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            coo_mapper_crs = coo_trafo.transform(_coordinates_crs_s)
        else:
            coo_mapper_crs = _coordinates_crs_s * 1.0

        pixel_row, pixel_col = self.coordinate_to_pixel(x=coo_mapper_crs[:, 0], y=coo_mapper_crs[:, 1])

        # Check if all pixels are valid
        # Even if the original pixel location is 0.3,0.3 -> anyhow a bilinear transformation is not possible because
        # Than the four points would be [-0.5,-0.5],[0.5,0.5],[-0.5,0.5],[0.5,-0.5] which is not possible
        # So the shifted pixel is fine to be not valid
        valid_mask = self.pixel_valid(px_row=pixel_row, px_col=pixel_col)
        valid_index = np.flatnonzero(valid_mask)
        issue = set()
        if not np.all(valid_mask):
            issue = {Issue.OUTSIDE_RASTER}
            if not np.any(valid_mask):
                return ResultFailure(
                    ok=False,
                    error="None of the coordinates are on the raster",
                    issues={Issue.OUTSIDE_RASTER},
                )

        # use only valid pixels
        pixel_row = pixel_row[valid_mask]
        pixel_col = pixel_col[valid_mask]

        # As the value of the raster point is basically defined for the center
        # we shift our target location by 0.5 for the bilinear interpolation and
        # say raster values are defined for the integer raster location
        # that we do not have to deal finding the right cells involved. By shifting by 0.5 the involved cells are
        # the ones we get from floor and ceil, if we say the values are defined not in the center but in the edge points

        pixel_row -= 0.5
        pixel_col -= 0.5

        # edge cases height= 100 -> max row index for array values = 99
        # width=200, max col = 199 -> max column index for array values
        # lower than 0.0 can no pixel be as we're then already raising an Error
        # row 0.1 ; col 100.2 ->  0,1; 100,101
        # row 98.9 ; col 100.2 ->  98,99; 100,101
        # -> row 99 ; col 100.2 ->  98,99; 100,101 need to check if floor(row/col) is height-1/width-1

        # Gdal does treats for pixels from 0 to 0.5 just as 0.5
        pixel_row[pixel_row < 0] = 0.0
        pixel_col[pixel_col < 0] = 0.0

        pixel_row[pixel_row > (self.height - 1)] = self.height - 1
        pixel_col[pixel_col > (self.width - 1)] = self.width - 1

        # TODO Could there be numerical problems? for example if its -0.0000000001
        # Should we treat that as still valid?
        row_lower = np.floor(pixel_row).astype(int)
        row_upper = row_lower * 1
        row_upper[row_lower < self.height - 1] += 1
        # if row_upper is max pixel size then row_upper is that already
        # Bigger than self.height is not possible because than it would already not valid
        row_lower[row_lower == self.height - 1] -= 1

        col_lower = np.floor(pixel_col).astype(int)
        col_upper = col_lower * 1
        col_upper[col_lower < self.width - 1] += 1
        # if row_upper is max pixel size then row_upper is that already
        col_lower[col_lower == self.width - 1] -= 1

        # value of that pixels
        r1 = self._raster_array[row_lower, col_lower]
        r2 = self._raster_array[row_upper, col_lower]
        r3 = self._raster_array[row_lower, col_upper]
        r4 = self._raster_array[row_upper, col_upper]

        # bilinear Interpolation for all points in the pixel system of the array
        # Actually we could shift the pixel to zero, but I think the numeric range of
        # pixel dimension (grid size) is small enough to not run into numeric problems.

        for _index, _ in enumerate(row_lower):
            cell_corners = [
                [row_lower[_index], col_lower[_index], r1[_index]],
                [row_upper[_index], col_lower[_index], r2[_index]],
                [row_lower[_index], col_upper[_index], r3[_index]],
                [row_upper[_index], col_upper[_index], r4[_index]],
            ]

            value = bilinear_interpolation(points=cell_corners, x=pixel_row[_index], y=pixel_col[_index])

            coo_mapper_crs[valid_index[_index], 2] = value

        coo_source_crs = np.empty(coo_mapper_crs.shape, dtype=float)
        coo_source_crs.fill(np.nan)
        # Transform back to source crs.
        if coo_trafo is not None:
            coo_source_crs[valid_index, :] = coo_trafo.transform(coo_mapper_crs[valid_index, :], direction="inverse")
        else:
            coo_source_crs[valid_index, :] = coo_mapper_crs[valid_index, :] * 1.0

        return MappingResultSuccess(
            ok=True,
            coordinates=coo_source_crs,
            mask=valid_mask,
            crs=crs_s,
            issues=issue,
        )



import pytest

from weitsicht.image.image_dict_selector import get_image_from_dict


def test_get_image_from_dict():
    # wrong dictionary
    with pytest.raises(KeyError):
        get_image_from_dict({})

    with pytest.raises(ValueError):
        get_image_from_dict({"type": "fisheye"})

    cam = {
        "type": "OpenCV",
        "calib_width": 1000,
        "calib_height": 600,
        "fx": 1000.0,
        "fy": 1000,
    }
    dict_load = {
        "type": "perspective",
        "width": 1000,
        "height": 600,
        "position": [1000.0, 2000.0, 0.0],
        "orientation_matrix": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
        "camera": cam,
    }

    image = get_image_from_dict(dict_load)

    assert image is not None

    dict_load_ortho = {
        "type": "ortho",
        "width": 1667,
        "height": 1667,
        "geo_transform": {
            "a": 0.29994001199760045,
            "b": 0.0,
            "c": -1750.0,
            "d": 0.0,
            "e": -0.29994001199760045,
            "f": 338800.0,
            "g": 0.0,
            "h": 0.0,
            "i": 1.0,
        },
        "resolution": 0.29994001199760045,
        "crs": None,
    }
    image = get_image_from_dict(dict_load_ortho)
    assert image is not None



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Perspective image model backed by a camera perspective implementation."""

from __future__ import annotations

import logging

import numpy as np
import pyproj.exceptions
from pyproj import CRS
from shapely import geometry

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.camera_dict_selector import get_camera_from_dict
from weitsicht.exceptions import (
    CRSInputError,
    MapperMissingError,
    MappingBackendError,
    NotGeoreferencedError,
    WeitsichtError,
)
from weitsicht.image.base_class import ImageBase, ImageType
from weitsicht.mapping.base_class import MappingBase
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.transform.rotation import Rotation
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    Issue,
    MappingResult,
    MappingResultSuccess,
    MaskN_,
    ProjectionResult,
    ProjectionResultSuccess,
    ResultFailure,
    Vector3D,
    to_array_nx2,
    to_array_nx3,
)

__all__ = ["ImagePerspective"]

logger = logging.getLogger(__name__)


class ImagePerspective(ImageBase):
    """Perspective image model.

    The image is geo-referenced if position, orientation, camera and CRS are specified.
    The camera model implementation should support resampled images (i.e. account for
    calibration size vs. current image size).
    """

    def __init__(
        self,
        width: float | int,
        height: float | int,
        camera: CameraBasePerspective | None,
        crs: CRS | None = None,
        position: Vector3D | None = None,
        orientation: Rotation | None = None,
        mapper: MappingBase | None = None,
    ):
        """Initialize a perspective image model.

        :param width: Image width in pixels.
        :type width: float | int
        :param height: Image height in pixels.
        :type height: float | int
        :param camera: Camera model used for projection/distortion, may be ``None`` for non-geo-referenced instances.
        :type camera: CameraBasePerspective | None
        :param crs: World CRS of the image, defaults to ``None``.
        :type crs: CRS | None
        :param position: Camera position in world CRS, defaults to ``None``.
        :type position: Vector3D | None
        :param orientation: Camera orientation, defaults to ``None``.
        :type orientation: Rotation | None
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        """

        super().__init__(mapper=mapper, crs=crs, width=width, height=height)

        self._position: Vector3D | None = position
        self._orientation: Rotation | None = orientation

        self._camera = camera

        # TODO check if crs is not None that the CRS provides a vertical reference

    @property
    def type(self) -> ImageType:
        """Return the image model type.

        :return: Image type.
        :rtype: ImageType
        """
        return ImageType.Perspective

    @classmethod
    def from_dict(cls, param_dict: dict, mapper: MappingBase | None = None) -> ImagePerspective:
        """Create an :class:`ImagePerspective` from a parameter dictionary.

        Required keys are:
        - ``width``
        - ``height``

        Optional keys are:
        - ``position`` (3D)
        - ``orientation_matrix`` (3x3)
        - ``crs`` (WKT)
        - ``camera`` (camera ``param_dict``)

        :param param_dict: Dictionary with image parameters.
        :type param_dict: dict
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        :return: Image model instance.
        :rtype: ImagePerspective
        :raises KeyError: If required keys are missing.
        :raises ValueError: If values are invalid.
        :raises TypeError: If values have incompatible types.
        :raises CRSInputError: If the CRS WKT string is invalid or unsupported.
        """

        width = param_dict["width"]
        height = param_dict["height"]

        if width == 0 or height == 0:
            raise ValueError("Image width and height can not be 0")

        position = None
        if param_dict.get("position") is not None:
            position = np.array(param_dict.get("position"))

        orientation = None
        if param_dict.get("orientation_matrix") is not None:
            r_matrix = np.array(param_dict.get("orientation_matrix"))
            orientation = Rotation(rotation_matrix=r_matrix)

        crs = None
        if param_dict.get("crs") is not None:
            try:
                crs = CRS(param_dict["crs"])
            except pyproj.exceptions.CRSError as err:
                raise CRSInputError("No valid CRS wkt string supported") from err

        camera = None
        if param_dict.get("camera") is not None:
            camera = get_camera_from_dict(param_dict["camera"])

        image = cls(
            width=width,
            height=height,
            camera=camera,
            position=position,
            orientation=orientation,
            crs=crs,
            mapper=mapper,
        )

        return image

    @property
    def param_dict(self) -> dict:
        """Return image parameters as a dictionary.

        The returned dictionary is compatible with :meth:`from_dict`.

        :return: Image parameters.
        :rtype: dict
        """
        param_dict = {
            "type": ImageType.Perspective.fullname,
            "width": self.width,
            "height": self.height,
            "position": self._position.tolist() if self._position is not None else None,
            "orientation_matrix": self._orientation.matrix.tolist() if self._orientation is not None else None,
            "camera": self.camera.param_dict if self.camera is not None else None,
            "crs": self.crs_wkt,
        }

        return param_dict

    @property
    def is_geo_referenced(self) -> bool:
        """Return whether the image is geo-referenced.

        :return: ``True`` if position, orientation and camera are set, otherwise ``False``.
        :rtype: bool
        """
        if self._position is not None and self._orientation is not None and self._camera is not None:
            return True
        return False

    @property
    def camera(self):
        """Return the camera model used by this image.

        :return: Camera model instance or ``None``.
        :rtype: CameraBasePerspective | None
        """
        return self._camera

    @property
    def center(self) -> tuple[float, float]:
        """Return the image center used for mapping/projection.

        For perspective images this is the principal point as defined by the camera model.

        :return: Center point (x, y) in image pixel coordinates.
        :rtype: tuple[float, float]
        :raises NotGeoreferencedError: If no camera model is specified.
        """

        if self._camera is None:
            raise NotGeoreferencedError("camera is not specified")
        center = self._camera.principal_point(self.shape)
        return float(center[0]), float(center[1])

    def image_points_inside(self, point_image_coordinates: ArrayNx2) -> MaskN_:
        """Return whether undistorted image points are valid for this camera model.

        Within the camera class the distortion border is generated during initialization and
        is used to determine where distortion/undistortion is valid.

        :param point_image_coordinates: Image pixel coordinates (image CRS).
        :type point_image_coordinates: ArrayNx2
        :return: Boolean mask of shape ``(N,)``.
        :rtype: ``MaskN_``
        :raises NotGeoreferencedError: If no camera model is specified.
        """

        if self._camera is None:
            raise NotGeoreferencedError("camera is not specified")
        return self._camera.undistorted_image_points_inside(point_image_coordinates, self.shape)

        # old version using discarded parameter outer dist
        # min_border = np.logical_and(point_image_coordinates[:, 0] >= -outer_dist,
        #                            point_image_coordinates[:, 1] >= -outer_dist)
        # max_border = np.logical_and(point_image_coordinates[:, 0] < self.width + outer_dist,
        #                            point_image_coordinates[:, 1] < self.height + outer_dist)
        # valid_index = np.where(np.logical_and(min_border, max_border))[0]
        #
        # if not valid_index.size:
        #    return None
        # return valid_index

    def position_to_crs(self, crs_t: CRS) -> Vector3D | None:
        """Return the camera position in the target CRS.

        :param crs_t: Target CRS.
        :type crs_t: CRS
        :return: Position in ``crs_t`` or ``None`` if unavailable.
        :rtype: Vector3D | None
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """
        if self._position is not None and self._crs is not None:
            coo_trafo = CoordinateTransformer.from_crs(crs_s=self._crs, crs_t=crs_t)

            if coo_trafo is not None:
                coordinates_crs_t = coo_trafo.transform(self._position)
            else:
                coordinates_crs_t = self._position * 1.0

            return coordinates_crs_t[0, :]
        return None

    def project(
        self,
        coordinates: ArrayNx3,
        crs_s: CRS | None = None,
        transformer: CoordinateTransformer | None = None,
        to_distorted: bool = True,
    ) -> ProjectionResult:
        """Project 3D coordinates into image pixel coordinates.

        The projection is first computed in undistorted image coordinates. If ``to_distorted`` is ``True``,
        only pixels within the camera model's valid undistortion border are distorted; invalid pixels remain
        undistorted and are marked in the returned mask.

        :param coordinates: 3D coordinates to project.
        :type coordinates: ArrayNx3
        :param crs_s: CRS of the input coordinates, defaults to ``None``.
        :type crs_s: CRS | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :param to_distorted: Whether to return distorted image coordinates, defaults to ``True``.
        :type to_distorted: bool
        :return: Projection result.
        :rtype: ProjectionResult
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises CRSInputError: If both ``crs_s`` and ``transformer`` are provided.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        points_world_crs = to_array_nx3(coordinates)

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        # Transformer
        if crs_s is not None and transformer is not None:
            raise CRSInputError("Either crs or transformation can be used or both None")
        coo_trafo = (
            transformer if transformer is not None else CoordinateTransformer.from_crs(crs_s=crs_s, crs_t=self._crs)
        )

        if coo_trafo is not None:
            coordinates_image_crs = coo_trafo.transform(points_world_crs)
        else:
            coordinates_image_crs = points_world_crs * 1.0

        assert self._orientation is not None
        assert self._position is not None
        # Bring 3d points into Camera CRS by using rotation matrix and projection center
        pts_camera_crs = np.matmul(self._orientation.matrix.T, (coordinates_image_crs - self._position).T).T

        assert self._camera is not None
        pts_image_crs = self._camera.pts_camara_crs_to_image_pixel(pts_camera_crs, self.shape, to_distorted=False)

        # Here we check if the undistorted image point is inside the border for which distorted points are inside
        # This is important as non-linear function of distortion,
        # would map extreme far points back to the image due to the higher order polynomials
        #
        valid_pixel = self._camera.undistorted_image_points_inside(pts_image_crs, image_size=self.shape)

        # Only the valid pixel which are inside the distortion border are distorted
        pts_image_crs[valid_pixel, :] = self._camera.pts_camara_crs_to_image_pixel(
            pts_camera_crs[valid_pixel, :], self.shape, to_distorted=to_distorted
        )

        issue = set()
        if not np.all(valid_pixel):
            issue = {Issue.INVALID_PROJECTIIONS}
            if not np.any(valid_pixel):
                return ResultFailure(ok=False, error="None of the projections is valid", issues=issue)

        # return pts_image_crs, valid_pixel
        return ProjectionResultSuccess(ok=True, pixels=pts_image_crs, mask=valid_pixel, issues=issue)

    def pixel_to_ray_vector(self, pixel_pos: ArrayNx2, is_undistorted: bool = False) -> ArrayNx3:
        """Calculate world ray direction vectors for image pixels.

        :param pixel_pos: Image pixel coordinates (pixel CRS).
        :type pixel_pos: ArrayNx2
        :param is_undistorted: Whether input pixels are already undistorted, defaults to ``False``.
        :type is_undistorted: bool
        :return: Unit direction vectors in world CRS.
        :rtype: ArrayNx3
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        """

        _pixel_pos = to_array_nx2(pixel_pos)

        if not self.is_geo_referenced:
            raise NotGeoreferencedError("Image is not georeferenced")

        assert self._camera is not None
        pts_camera_crs = self._camera.pixel_image_to_camera_crs(_pixel_pos, self.shape, is_undistorted)

        assert self._orientation is not None
        line_vec = np.matmul(self._orientation.matrix, pts_camera_crs.T).T
        line_vec = np.divide(line_vec.T, np.linalg.norm(line_vec, axis=1)).T

        return line_vec

    def map_center_point(
        self, mapper: MappingBase | None = None, transformer: CoordinateTransformer | None = None
    ) -> MappingResult:
        """Map the image center pixel (principal point) to 3D coordinates.

        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises MappingBackendError: If the mapping backend fails unexpectedly.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_center_point")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        ray_vector = self.pixel_to_ray_vector(np.array(self.center))

        # checked pos and camera via is_geo_referenced
        assert mapper_to_use is not None
        assert self._position is not None
        assert self._camera is not None

        try:
            result_mapper = mapper_to_use.map_coordinates_from_rays(
                ray_vector,
                np.array([self._position]),
                None if transformer is not None else self._crs,
                transformer=transformer,
            )
        except WeitsichtError:
            raise
        except Exception as err:
            raise MappingBackendError(f"{type(mapper_to_use).__name__}.map_coordinates_from_rays failed") from err

        if result_mapper.ok is False:
            return ResultFailure(ok=False, error=result_mapper.error, issues=result_mapper.issues)

        coordinates = result_mapper.coordinates
        dist = np.linalg.norm(self._position - coordinates)
        gsd = float(dist / self._camera.focal_length_for_gsd_in_pixel)
        return MappingResultSuccess(
            ok=True,
            coordinates=coordinates,
            mask=result_mapper.mask,
            crs=self._crs,
            issues=result_mapper.issues,
            gsd=gsd,
        )

    def map_footprint(
        self,
        points_per_edge: int = 0,
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map the image footprint polygon to 3D coordinates.

        :param points_per_edge: Number of points inserted between corners, defaults to ``0``.
            Values ``< 1`` map only the corner points.
        :type points_per_edge: int
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises MappingBackendError: If the mapping backend fails unexpectedly.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_footprint")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        assert mapper_to_use is not None
        assert self._position is not None
        assert self._camera is not None

        if points_per_edge < 1:
            footprint_points_2d = np.array([[0, 0], [self.width, 0], [self.width, self.height], [0, self.height]])
        else:
            footprint_points_2d = [[x, 0] for x in np.linspace(0, self.width, 2 + points_per_edge)]
            footprint_points_2d += [[self.width, x] for x in np.linspace(0, self.height, 2 + points_per_edge)][1:]
            footprint_points_2d += [[x, self.height] for x in np.linspace(self.width, 0, 2 + points_per_edge)][1:]
            footprint_points_2d += [[0, x] for x in np.linspace(self.height, 0, 2 + points_per_edge)][1:-1]

        footprint_points_2d = np.array(footprint_points_2d)
        ray_vector = self.pixel_to_ray_vector(footprint_points_2d)

        ray_pos = np.ones(ray_vector.shape) * self._position

        try:
            result_mapper = mapper_to_use.map_coordinates_from_rays(
                ray_vector, ray_pos, None if transformer is not None else self._crs, transformer=transformer
            )
        except WeitsichtError:
            raise
        except Exception as err:
            raise MappingBackendError(f"{type(mapper_to_use).__name__}.map_coordinates_from_rays failed") from err

        if result_mapper.ok is False:
            return ResultFailure(ok=False, error=result_mapper.error, issues=result_mapper.issues)

        if not np.all(result_mapper.mask):
            return ResultFailure(
                ok=False,
                error="Some footprint points failed",
                issues=result_mapper.issues,
            )

        footprint_points_3d = result_mapper.coordinates

        footprint_geom = geometry.Polygon(footprint_points_3d)
        area = float(np.round(footprint_geom.area))
        # gsd = float(np.round(np.sqrt(area / (self.width * self.height)), 4))

        dist = np.linalg.norm(self._position - footprint_points_3d, axis=1)
        gsd = float(np.mean(dist / self._camera.focal_length_for_gsd_in_pixel))

        return MappingResultSuccess(
            ok=True,
            coordinates=footprint_points_3d,
            mask=result_mapper.mask,
            crs=self._crs,
            gsd=gsd,
            area=area,
            issues=result_mapper.issues,
        )

    def map_points(
        self,
        points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int],
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
        is_undistorted: bool = False,
    ) -> MappingResult:
        """Map image pixel coordinates to 3D coordinates via a mapper.

        It is also possible to pass undistorted pixels. This can be useful if you already computed
        undistorted pixel coordinates from the original distorted image content.

        :param points_image: Pixel coordinates (distorted or undistorted).
        :type points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int]
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`~weitsicht.image.base_class.ImageBase.mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :param is_undistorted: Whether input pixels are already undistorted, defaults to ``False``.
        :type is_undistorted: bool
        :return: Mapping result.
        :rtype: MappingResult
        :raises ValueError: If ``points_image`` cannot be parsed as an array of 2D points.
        :raises NotGeoreferencedError: If the image is not geo-referenced.
        :raises MapperMissingError: If no mapper is available.
        :raises CRSInputError: If the mapper rejects CRS/transformer input.
        :raises CoordinateTransformationError: If coordinate transformation fails.
        :raises MappingBackendError: If the mapping backend fails unexpectedly.
        :raises WeitsichtError: Base class for all weitsicht exceptions. Catch this to handle any weitsicht error;
            catch specific subclasses first if you need to distinguish causes.
        """
        # TODO what should we do if specified points_image are outside the image dimensions
        # but the mapping still works? Should be masked them as invalid?
        # different to the orthoimage here far outside pixel coordinates could be distorted back
        # to valid areas.

        if not self.is_geo_referenced:
            raise NotGeoreferencedError()

        if self.mapper is None and mapper is None:
            raise MapperMissingError("No mapper provided for map_points")

        mapper_to_use = self.mapper
        if mapper is not None:
            mapper_to_use = mapper

        try:
            _points_image = to_array_nx2(points_image)
        except (IndexError, TypeError, ValueError) as err:
            raise ValueError("points_image must be array-like with shape (N, 2) or (N, 3)") from err
        # build ray vectors of image points
        ray_vector = self.pixel_to_ray_vector(_points_image, is_undistorted=is_undistorted)
        # build ray start points for all rays
        assert self._position is not None
        ray_pos = np.ones(ray_vector.shape) * self._position

        assert mapper_to_use is not None
        # Map points using either the images specified mapper or the user mapper provided

        try:
            mapping_result: MappingResult = mapper_to_use.map_coordinates_from_rays(
                ray_vector, ray_pos, None if transformer is not None else self._crs, transformer=transformer
            )
        except WeitsichtError:
            raise
        except Exception as err:
            raise MappingBackendError(f"{type(mapper_to_use).__name__}.map_coordinates_from_rays failed") from err

        if mapping_result.ok is False:
            return ResultFailure(ok=False, error=mapping_result.error, issues=mapping_result.issues)

        assert self._camera is not None
        dist = np.mean(np.linalg.norm(self._position - mapping_result.coordinates, axis=1))
        gsd = dist / self._camera.focal_length_for_gsd_in_pixel

        # TODO Implementation of gsd per point, optional it could be already in the mapper using normals at points
        return MappingResultSuccess(
            ok=True,
            coordinates=mapping_result.coordinates,
            mask=mapping_result.mask,
            gsd=gsd,
            issues=mapping_result.issues,
        )



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


"""Exceptions used in the package.

Design goal:
- Provide a small, well-defined hierarchy so users can either catch one base class
or handle specific subclasses.
"""


class WeitsichtError(Exception):
    """Base exception for the weitsicht package."""


class NotGeoreferencedError(WeitsichtError):
    """No valid georeference is provided for images"""


class CRSInputError(WeitsichtError, ValueError):
    """Invalid CRS/transformer input was provided."""


class CRSnoZaxisError(CRSInputError):
    """CRS does not define a vertical (Z) axis."""


class CoordinateTransformationError(WeitsichtError):
    """Coordinate Conversion is not possible"""


class MappingError(WeitsichtError):
    """Mapping of the geometry was not possible"""


class MappingBackendError(MappingError):
    """A mapping backend raised an unexpected/foreign exception."""


class MapperMissingError(MappingError):
    """No valid mapper is provided for image mapping functions"""



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Test of ortho photos"""

from pathlib import Path

import numpy as np
import pyproj.network
import pytest
from pyproj import CRS

from weitsicht.image.base_class import ImageType
from weitsicht.image.ortho import ImageOrtho
from weitsicht.mapping.raster import MappingRaster
from weitsicht.utils import Issue

DATA_DIR = Path(__file__).parent.resolve() / "data"
pyproj.network.set_network_enabled(True)  # type: ignore


@pytest.fixture
def image():
    image = ImageOrtho.from_file(DATA_DIR / "44_2_op_2023_30cm.tif", CRS("EPSG:31256+5778"))

    assert image is not None
    assert int(image.type)

    assert image.position_wgs84 is not None
    assert image.position_wgs84_geojson is not None

    assert image.crs_wkt is not None
    assert image.crs_proj4 is not None

    image.mapper = MappingRaster(DATA_DIR / "44_2_dgm.tif", crs=CRS("EPSG:31256+8881"))

    assert np.isclose(image.resolution, 0.3, atol=0.001, rtol=0)

    assert image.geo_transform is not None

    return image


def test_type(image):
    assert image.type == ImageType.Orthophoto


def test_from_dict():

    crs = CRS("EPSG:31256+5778")
    dict_load = {
        "type": "ortho",
        "width": 1667,
        "height": 1667,
        "geo_transform": {
            "a": 0.29994001199760045,
            "b": 0.0,
            "c": -1750.0,
            "d": 0.0,
            "e": -0.29994001199760045,
            "f": 338800.0,
            "g": 0.0,
            "h": 0.0,
            "i": 1.0,
        },
        "resolution": 0.29994001199760045,
        "crs": crs.to_wkt(),
    }

    image_from_dict = ImageOrtho.from_dict(dict_load)
    coo = np.array(
        [
            [-1529.31, 338396.05, 34.7672685000015],
            [-1500.38, 338385.504, 34.7062945599976],
        ]
    )
    pix = np.array([[735.78046, 1346.7693], [832.23308, 1381.929664]])

    assert image_from_dict is not None
    image_from_dict.mapper = MappingRaster(DATA_DIR / "44_2_dgm.tif", crs=CRS("EPSG:31256+8881"))

    result = image_from_dict.map_points(pix)

    # Vienna height to GHA is a simple shift, thus, we are also testing the transformer here
    coo_gha = coo + np.array([0, 0, 156.68])
    assert result.ok is True
    assert np.allclose(result.coordinates, coo_gha, atol=1e-7, rtol=0)


def test_param_dict(image):
    test_dict = {
        "type": "ortho",
        "width": 1667,
        "height": 1667,
        "geo_transform": {
            "a": 0.29994001199760045,
            "b": 0.0,
            "c": -1750.0,
            "d": 0.0,
            "e": -0.29994001199760045,
            "f": 338800.0,
            "g": 0.0,
            "h": 0.0,
            "i": 1.0,
        },
        "resolution": 0.29994001199760045,
        "crs": image.crs_wkt,
    }

    assert image.param_dict == test_dict


def test_center(image):
    assert image.center == (833.5, 833.5)


def test_is_geo_referenced(image):
    assert image.is_geo_referenced


def test_image_points_inside(image):
    assert np.all(image.image_points_inside(np.array([[0, 0], [1200, 1200]])))
    assert not np.all(image.image_points_inside(np.array([[-1, -1], [1667, 1200]])))


def test_position_to_crs(image):
    # For orthophoto the image is the center of image using the mapper if specified for height
    # otherwise its zero in the images coordinates

    # Using GDAL for the center
    # gdallocationinfo - r bilinear - geoloc 44_2_dgm.tif - 1.5000e+033.3855e+05
    # Value: 33.17

    result = image.position_to_crs(CRS("EPSG:31256+5778").to_3d())
    assert np.isclose(result[2], 33.17 + 156.68, atol=0.0001, rtol=0.0)

    # Test where trafo of mapper and code parts which use transformation
    result = image.position_to_crs(image.mapper.crs)
    assert np.isclose(result[2], 33.17, atol=0.0001, rtol=0.0)

    # Case no mapper is specified
    image.mapper = None
    result = image.position_to_crs("EPSG:31256+5778")
    assert np.isclose(result[2], 0, atol=0.0001, rtol=0.0)


def test_project(image):
    # gdallocationinfo -r bilinear -geoloc 44_2_op_2023_30cm.tif -1529.31 338396.05
    #   Location: (735.78046P,1346.76930000004L)
    # gdallocationinfo -r bilinear -geoloc 44_2_op_2023_30cm.tif -1500.38 338385.504
    #   Location: (832.23308P,1381.92966399994L)

    ref_gdal = np.array([[735.78046, 1346.76930000004], [832.23308, 1381.92966399994]])

    coo = np.array([[-1529.31, 338396.05], [-1500.38, 338385.504]])

    result = image.project(coo, image.crs)

    assert result.ok is True
    np.testing.assert_array_equal(result.mask, np.array([True, True]))
    assert np.allclose(result.pixels, ref_gdal, atol=1e-8, rtol=0)

    coo = np.array([[-1760, 338396.05], [-1500.38, 338809.504]])
    result = image.project(coo, image.crs)
    assert result.ok is False
    assert Issue.INVALID_PROJECTIIONS in result.issues

    coo = np.array(
        [
            [-1529.31, 338396.05],
            [-1760, 338396.05],
            [-1500.38, 338809.504],
            [-1500.38, 338385.504],
        ]
    )
    result = image.project(coo, image.crs)
    assert result.ok is True
    assert Issue.INVALID_PROJECTIIONS in result.issues
    np.testing.assert_array_equal(result.mask, np.array([True, False, False, True]))


def test_map_center_point(image):
    # Using GDAL for the center
    # gdallocationinfo - r bilinear - geoloc 44_2_dgm.tif - 1.5000e+033.3855e+05
    # Value: 33.17
    # that value is in vienna height. Shift by 156.68 to be in GHA heights of image

    result = image.map_center_point()
    assert result.ok is True
    assert np.isclose(result.coordinates[0][2], 33.17 + 156.68, atol=0.0001, rtol=0.0)


def test_map_footprint(image):

    result = image.map_footprint()
    assert result.ok is True
    assert np.isclose(result.area, 250000, atol=1e-1, rtol=0)
    assert np.isclose(result.gsd, 0.3, atol=1e-1, rtol=0)

    # gdallocationinfo -r bilinear -geoloc 44_2_dgm.tif  -1529.31 338800.0 -> 26.036
    # gdallocationinfo -r bilinear -geoloc 44_2_dgm.tif  -1750.0  338300.0 -> 35.982
    # gdallocationinfo -r bilinear -geoloc 44_2_dgm.tif  -1250.0  338300.0 -> 33.887
    # gdallocationinfo -r bilinear -geoloc 44_2_dgm.tif  -1250.0  338800.0 -> 33.517

    ref_border = np.array(
        [
            [-1750.0, 338800.0, 26.036],
            [-1750.0, 338300.0, 35.982],
            [-1250.0, 338300.0, 33.887],
            [-1250.0, 338800.0, 33.517],
        ]
    )
    # shift to gha
    ref_border = ref_border + np.array([0, 0, 156.68])

    assert np.allclose(ref_border, result.coordinates, atol=1e-6, rtol=0)


def test_map_points(image):
    # Heights from GDAL
    # gdallocationinfo -r bilinear -geoloc 44_2_dgm.tif  -1529.31 338396.05
    # Location: (231.19P, 414.450000000012L)
    # Value: 34.7672685000015

    # gdallocationinfo - r bilinear - geoloc 44_2_dgm.tif - 1500.38 338385.504
    # Location: (260.12P, 424.995999999985L)
    # Value: 34.7062945599976

    coo = np.array(
        [
            [-1529.31, 338396.05, 34.7672685000015],
            [-1500.38, 338385.504, 34.7062945599976],
        ]
    )
    pix = np.array([[735.78046, 1346.7693], [832.23308, 1381.929664]])

    result = image.map_points(pix)

    # Vienna height to GHA is a simple shift, thus we are also testing the transformer here
    coo_gha = coo + np.array([0, 0, 156.68])
    assert result.ok is True
    assert np.allclose(result.coordinates, coo_gha, atol=1e-7, rtol=0)

    # Test outside Orthophoto
    # but inside mapping Raster
    # Test that we are inside the mapping raster
    # Not sure why I have tested this
    # res = image.mapper.map_heights_from_coordinates(np.array([[-1760, 338396.05]]), image.mapper.crs)
    # assert res is not None

    pix = np.array([[0, 0], [-1, -1], [735.78046, 1346.7693], [832.23308, 1381.929664]])
    result = image.map_points(pix)
    assert result.ok is True
    np.testing.assert_array_equal(result.mask, np.array([True, True, True, True]))

    # TODO find a test where some points are outside the mapping raster to check that issues are returned
    # assert Issue.OUTSIDE_RASTER in result.issues



import numpy as np
import pyproj.network
from pyproj import CRS, Transformer

from weitsicht.transform.coordinates_transformer import CoordinateTransformer

pyproj.network.set_network_enabled(True)  # pyright: ignore[reportPrivateImportUsage]


def test_from_pipeline():
    # Need to do that. Need proper pipeline for testing
    assert True


def test_transform():
    # MGI Lambert in Bessel Ellipsoid
    ref_p_1_mgi_lambert_ell = np.array([556538.9876, 429158.5652, 2000])
    # from BEV transformator https://transformator.bev.gv.at/at.gv.bev.transformator/austrian
    # ref_p_1_mgi_9266 = np.array([4142953.6487, 1142861.1838, 4698745.5662])
    ref_p1_mgi_9267_bessel_ell = np.array([15.42187111, 47.743444717, 2000.000])

    # lambert with ellipsoid (bessel) height to MGI
    # Directly using pyproj Transformer
    t_0_1 = Transformer.from_crs(CRS(31287).to_3d(), 9267, always_xy=True, allow_ballpark=False, only_best=True)
    p_0_1 = np.array(
        t_0_1.transform(
            ref_p_1_mgi_lambert_ell[0],
            ref_p_1_mgi_lambert_ell[1],
            ref_p_1_mgi_lambert_ell[2],
        )
    )
    assert np.allclose(p_0_1, ref_p1_mgi_9267_bessel_ell, atol=0.001, rtol=0)
    # Using CoordinateTransformer class
    c_0_1 = CoordinateTransformer.from_crs(CRS(31287).to_3d(), CRS(9267))
    assert c_0_1 is not None
    p_0_1_c = c_0_1.transform(ref_p_1_mgi_lambert_ell)
    assert np.allclose(p_0_1_c, ref_p1_mgi_9267_bessel_ell, atol=1e-8, rtol=0)
    # Backwards
    p_0_1_c_inverse = c_0_1.transform(p_0_1_c, direction="inverse")
    assert p_0_1_c_inverse is not None
    assert np.allclose(p_0_1_c_inverse, ref_p_1_mgi_lambert_ell, atol=0.001, rtol=0)

    # USING national height system (grids involved)
    # MGI Lambert in gha
    ref_p_1_mgi_lambert_gha = np.array([556538.9876, 429158.5652, 2000])
    # from BEV transformator https://transformator.bev.gv.at/at.gv.bev.transformator/austrian
    ref_p1_mgi_4312_gha = np.array([15.421871111, 47.743444716, 2000.000])

    # lambert with gha heights to MGI (lat,lon) gha
    crs_lambert_mgi_gha = CRS("EPSG:31287+9274")
    # 9267 is geog3d with ell height. We need to combine 2d ell with gha
    crs_mgi_latlon_gha = CRS("EPSG:4312+9274")

    # using pyproj Transformer
    t_0_2 = Transformer.from_crs(
        crs_lambert_mgi_gha,
        crs_mgi_latlon_gha,
        always_xy=True,
        allow_ballpark=False,
        only_best=True,
    )

    p_0_2 = t_0_2.transform(
        ref_p_1_mgi_lambert_gha[0],
        ref_p_1_mgi_lambert_gha[1],
        ref_p_1_mgi_lambert_gha[2],
    )
    assert np.allclose(p_0_2, ref_p1_mgi_4312_gha, atol=1e-9, rtol=0)

    # using CoordinateTransformer class
    c_0_2 = CoordinateTransformer.from_crs(crs_lambert_mgi_gha, crs_mgi_latlon_gha)
    assert c_0_2 is not None
    p_0_2_c = c_0_2.transform(ref_p_1_mgi_lambert_gha)
    assert np.allclose(p_0_2_c, ref_p1_mgi_4312_gha, atol=1e-8, rtol=0)

    # backwards
    assert c_0_2 is not None
    p_0_2_c_inverse = c_0_2.transform(ref_p1_mgi_4312_gha, direction="inverse")
    assert np.allclose(p_0_2_c_inverse, ref_p_1_mgi_lambert_gha, atol=1e-3, rtol=0)

    # (2) test to MGI GK and to UTM with different heights
    # Most references are taken from BEV transformator
    # BEV transformator https://transformator.bev.gv.at/at.gv.bev.transformator/austrian
    # If using pyproj it will get all transformations and epsg from a database of the current pyproj version
    # If nothing specified it will use the best transformation also containing v and h grids.

    # Using heights in use of Austria GHA (Gebrauchshoehen)
    # Lambert to GK MGI
    # That one works well
    pt_test_lambert = np.array([556538.9876, 429158.5652, 2000])
    # From BEV transformator
    ref_mgi_gk_31_gha = np.array([156606.740, 5291475.650, 2000.000])

    crs_lambert_mgi_gha = CRS("EPSG:31287+9274")
    crs_gk_31_gha = CRS("EPSG:9272+9274")
    c_1_1 = CoordinateTransformer.from_crs(crs_lambert_mgi_gha, crs_gk_31_gha)
    assert c_1_1 is not None
    p_1_1_c = c_1_1.transform(pt_test_lambert)
    assert np.allclose(ref_mgi_gk_31_gha, p_1_1_c, atol=0.001, rtol=0)
    # Backwards
    p_transformed_inverse = c_1_1.transform(ref_mgi_gk_31_gha, direction="inverse")
    assert np.allclose(pt_test_lambert, p_transformed_inverse, atol=0.001, rtol=0)

    # Using heights in use of Austria GHA (Gebrauchshoehen) for Lambert to UTM with ell heights
    # That one works well
    pt_test_lambert = np.array([556538.9876, 429158.5652, 2000])
    # From BEV transformator -  UTM ell heights grs80
    pt_ref_utm_ell = np.array([531547.645, 5287818.137, 2047.163])

    crs_lambert_mgi_gha = CRS("EPSG:31287+5778")
    crs_etrs_utm_ell = CRS("EPSG:25833").to_3d()  # We need to make utm 3d that its using 3d transformation correctly

    c_2_2 = CoordinateTransformer.from_crs(crs_lambert_mgi_gha, crs_etrs_utm_ell)
    assert c_2_2 is not None
    p_2_2_c = c_2_2.transform(pt_test_lambert)
    assert np.allclose(p_2_2_c, pt_ref_utm_ell, atol=0.001, rtol=0)

    # Backwards
    p_2_2_c_inverse = c_2_2.transform(pt_ref_utm_ell, direction="inverse")
    assert np.allclose(pt_test_lambert, p_2_2_c_inverse, atol=0.001, rtol=0)

    # Using heights in use of Austria GHA (Gebrauchshoehen) for Lambert to UTM with egm2008 heights
    # That one works well
    pt_test_lambert = np.array([556538.9876, 429158.5652, 2000])

    # From BEV transformator
    pt_ref_utm_ell = np.array([531547.645, 5287818.137, 2047.163])
    # Self calculated directly from UTM ell to UTM egm2008
    # We need to make UTM to 3d that its aware of ellipsoid heights
    c_utm_ell_utm_egm2008 = Transformer.from_crs(CRS("EPSG:25833").to_3d(), "EPSG:25833+3855")
    pt_ref_utm_egm2008 = np.array(c_utm_ell_utm_egm2008.transform(*pt_ref_utm_ell))

    crs_lambert_mgi_gha = CRS("EPSG:31287+5778")
    crs_etrs_utm_egm2008 = CRS("EPSG:25833+3855")
    c_2_3 = CoordinateTransformer.from_crs(crs_lambert_mgi_gha, crs_etrs_utm_egm2008)
    assert c_2_3 is not None
    p_2_3_c = c_2_3.transform(pt_test_lambert)
    assert np.allclose(p_2_3_c, pt_ref_utm_egm2008, atol=0.001, rtol=0)

    # Backwards
    p_2_3_c_inverse = c_2_3.transform(pt_ref_utm_egm2008, direction="inverse")
    assert np.allclose(pt_test_lambert, p_2_3_c_inverse, atol=0.001, rtol=0)

    # Now we check that case why directly using pyproj Transformer can make a problem
    # This is where the direct use of pyproj Transformer fails

    p_wgs84_egm2008 = np.array([15.420831277, 47.742966114, 2000])
    # To get reference from BEV we will transform first to ell coordinates which is correct here anyhow
    t_3_0 = Transformer.from_crs(CRS("EPSG:4326+3855"), 4979, always_xy=True)
    p_wgs84_ell = t_3_0.transform(*p_wgs84_egm2008)

    print(p_wgs84_ell)  # That we use in BEV transformator
    p_ref_lambert_gha = np.array([556539.049, 429158.797, 2000.478])

    crs_wgs84_egm2008 = CRS("EPSG:4326+3855")
    crs_mgi_gha = CRS("EPSG:31287+5778")
    t_3_1 = Transformer.from_crs(
        crs_wgs84_egm2008,
        crs_mgi_gha,
        always_xy=True,
        allow_ballpark=False,
        only_best=True,
    )
    p_3_1 = t_3_1.transform(*p_wgs84_egm2008)
    # That point is now wrong
    print("wrong:", p_3_1, " diff:", p_3_1 - p_ref_lambert_gha)

    c_3_1 = CoordinateTransformer.from_crs(crs_wgs84_egm2008, crs_mgi_gha)
    assert c_3_1 is not None
    p_3_1_c = c_3_1.transform(p_wgs84_egm2008)
    # Using our Coordinate Transformer it works
    print("correct:", p_3_1_c, " diff:", p_3_1_c - p_ref_lambert_gha)

    # WGS84/UTM also does not work directly
    p_wgs84_utm_egm2008 = np.array([531547.712, 5287818.367, 2000.000])
    # To get reference from BEV we will transform first to ell coordinates which is correct here anyhow
    p_ref_lambert_gha = np.array([556539.049, 429158.797, 2000.478])

    crs_wgs84_utm_egm2008 = CRS("EPSG:32633+3855")
    crs_mgi_gha = CRS("EPSG:31287+5778")
    t_3_2 = Transformer.from_crs(
        crs_wgs84_utm_egm2008,
        crs_mgi_gha,
        always_xy=True,
        allow_ballpark=False,
        only_best=True,
    )
    p_3_2 = t_3_2.transform(*p_wgs84_utm_egm2008)
    # That point is now wrong
    print("wrong:", p_3_2, " diff:", p_3_2 - p_ref_lambert_gha)

    c_3_2 = CoordinateTransformer.from_crs(crs_wgs84_utm_egm2008, crs_mgi_gha)
    assert c_3_2 is not None
    p_3_2_c = c_3_2.transform(p_wgs84_utm_egm2008)
    # Using our Coordinate Transformer it works
    print("correct:", p_3_2_c, " diff:", p_3_2_c - p_ref_lambert_gha)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Base types and interfaces for image models."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum

from pyproj import CRS

from weitsicht.geometry.coo_geojson import get_geojson
from weitsicht.mapping.base_class import MappingBase
from weitsicht.transform.coordinates_transformer import CoordinateTransformer
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    MappingResult,
    ProjectionResult,
    Vector3D,
)

__all__ = ["ImageType", "ImageBase"]


class ImageType(Enum):
    """Enum of supported image model types.

    The ``fullname`` attribute is used for serialization (e.g. in ``param_dict``)
    and must match what factory functions expect (e.g. ``get_image_from_dict``).
    Add new image model types here.
    """

    fullname: str
    Unknown = 0, "unknown"
    Perspective = 1, "perspective"
    Orthophoto = 2, "ortho"

    def __new__(cls, value, name):
        member = object.__new__(cls)
        member._value_ = value
        member.fullname = name
        return member

    def __int__(self):
        return self.value


class ImageBase:
    """Base class for image models.

    This base class defines a common API for projecting 3D coordinates into image
    pixel coordinates and mapping image pixels to 3D coordinates (via an optional
    :class:`~weitsicht.mapping.base_class.MappingBase`).

    Coordinate system conventions:
    - Image pixel CRS: x points right, y points down (origin at top-left pixel edge).
    - World CRS: any cartesian CRS; for mapping it is usually best to use a projected CRS.

    Subclasses must implement:
    - `from_dict`
    - `type`
    - `param_dict`
    - `is_geo_referenced`
    - `center`
    - `position_to_crs`
    - `project`
    - `map_center_point`
    - `map_footprint`
    - `map_points`
    """

    def __init__(
        self,
        width: float | int,
        height: float | int,
        mapper: MappingBase | None = None,
        crs: CRS | None = None,
    ):
        """Initialize the image base class.

        :param width: Image width in pixels.
        :type width: float | int
        :param height: Image height in pixels.
        :type height: float | int
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        :param crs: World CRS of the image, defaults to ``None``.
        :type crs: CRS | None
        """

        self._crs: CRS | None = crs
        self.mapper = mapper

        self._width: int = int(width)
        self._height: int = int(height)

        self._position = None

    @property
    def mapper(self) -> MappingBase | None:
        # get mapper
        return self._mapper

    @mapper.setter
    def mapper(self, mapper: MappingBase | None):
        self._mapper = mapper

    @classmethod
    @abstractmethod
    def from_dict(cls, param_dict: dict, mapper: MappingBase | None = None) -> ImageBase:
        """Create an image model from a parameter dictionary.

        Implementations should accept the dictionary returned by :attr:`param_dict`.

        :param param_dict: Dictionary with image parameters.
        :type param_dict: dict
        :param mapper: Mapping instance, defaults to ``None``.
        :type mapper: MappingBase | None
        :return: Image model instance.
        :rtype: ImageBase
        :raises KeyError: If required keys are missing.
        :raises ValueError: If configuration values are invalid.
        :raises TypeError: If configuration values have incompatible types.
        """
        pass

    @property
    @abstractmethod
    def type(self) -> ImageType:
        """Return the image model type.

        :return: Image type.
        :rtype: ImageType
        """
        pass

    @property
    @abstractmethod
    def param_dict(self) -> dict:
        """Return image parameters as a dictionary.

        The returned dictionary must be compatible with :meth:`from_dict`.

        :return: Image parameters.
        :rtype: dict
        """
        pass

    @property
    @abstractmethod
    def is_geo_referenced(self) -> bool:
        """Return whether the image is geo-referenced.

        :return: ``True`` if geo-referenced (and mapping/projection can be computed), otherwise ``False``.
        :rtype: bool
        """
        pass

    @property
    @abstractmethod
    def center(self) -> tuple[float, float]:
        """Return the center of the image in pixel coordinates.

        Implementations may define the center differently (e.g. the principal point
        for perspective cameras).

        :return: Center point (x, y) in pixel coordinates.
        :rtype: tuple[float, float]
        """
        pass

    @abstractmethod
    def position_to_crs(self, crs_t: CRS) -> Vector3D | None:
        """Return the image reference position in the target CRS.

        :param crs_t: Target CRS.
        :type crs_t: CRS
        :return: Position in ``crs_t`` or ``None`` if unavailable.
        :rtype: Vector3D | None
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """
        pass

    @abstractmethod
    def project(
        self, coordinates: ArrayNx3, crs_s: CRS | None = None, transformer: CoordinateTransformer | None = None
    ) -> ProjectionResult:
        """Project 3D coordinates into image pixel coordinates.

        :param coordinates: 3D coordinates to project.
        :type coordinates: ArrayNx3
        :param crs_s: CRS of the input coordinates, defaults to ``None``.
        :type crs_s: CRS | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Result of the projection.
        :rtype: ProjectionResult
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """
        pass

    @abstractmethod
    def map_center_point(
        self, mapper: MappingBase | None = None, transformer: CoordinateTransformer | None = None
    ) -> MappingResult:
        """Map the image center pixel to 3D coordinates.

        For perspective images this is typically the principal point.

        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """
        pass

    @abstractmethod
    def map_footprint(
        self,
        points_per_edge: int = 0,
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map footprint of image to 3D.

        :param points_per_edge: The number of points inserted between corners, defaults to ``0``.
            ``0`` means only corner points are mapped.
        :type points_per_edge: int
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        pass

    @abstractmethod
    def map_points(
        self,
        points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int],
        mapper: MappingBase | None = None,
        transformer: CoordinateTransformer | None = None,
    ) -> MappingResult:
        """Map image pixel coordinates to 3D coordinates.

        :param points_image: Pixel coordinates.
        :type points_image: ArrayNx2 | ArrayNx3 | list[list[float]] | list[list[int]] | list[float] | list[int]
        :param mapper: Mapper to use, defaults to ``None`` (uses :attr:`mapper`).
        :type mapper: MappingBase | None
        :param transformer: Optional coordinate transformer, defaults to ``None``.
        :type transformer: CoordinateTransformer | None
        :return: Mapping result.
        :rtype: MappingResult
        :raises CoordinateTransformationError: If coordinate transformation fails.
        """

        pass

    # Standard properties
    @property
    def position_wgs84(self) -> tuple[float, float, float] | None:
        """Return the position in WGS84 (EPSG:4979) of the image or None if not possible.
        The position_to_crs method is defined by the class implementation.

        :return: Position in WGS84 (x, y, z) or ``None``.
        :rtype: tuple[float, float, float] | None
        """

        coordinates = self.position_to_crs(CRS.from_epsg(4979))

        if coordinates is not None:
            pos_wgs84 = coordinates
            return float(pos_wgs84[0]), float(pos_wgs84[1]), float(pos_wgs84[2])

        return None

    @property
    def position_wgs84_geojson(self) -> dict | None:
        """Return the image position in WGS84 (EPSG:4979) as a GeoJSON point.

        :return: GeoJSON ``Point`` mapping or ``None`` if the position is unavailable.
        :rtype: dict | None
        """
        coordinates = self.position_to_crs(CRS.from_epsg(4979))
        if coordinates is not None:
            return get_geojson(coordinates, "Point")
        return None

    @property
    def crs(self) -> CRS | None:
        """Return the image world CRS.

        :return: CRS of the image or ``None`` if unknown.
        :rtype: CRS | None
        """
        return self._crs

    # TODO should we actually perform a transformation?
    # Then we would need orientation matrix to be transformed accordingly
    @crs.setter
    def crs(self, crs: CRS):
        """Set images CRS from pyproj CRS object

        :param crs: CRS of the image.
        :type crs: CRS
        """
        self._crs = crs

    @property
    def crs_wkt(self) -> str | None:
        """Return images CRS as wkt string"""
        if self._crs is not None:
            return self._crs.to_wkt()
        return None

    @property
    def crs_proj4(self) -> str | None:
        """Return images CRS as proj4 string"""
        if self._crs is not None:
            return self._crs.to_proj4()
        return None

    @property
    def width(self) -> int:
        """Return width of image (column number)"""
        return self._width

    @property
    def height(self) -> int:
        """Return height of image (row number)"""
        return self._height

    @property
    def shape(self) -> tuple[int, int]:
        """Return shape of image as tuple(width, height)"""
        return self._width, self._height



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

from pathlib import Path

import numpy as np
import pytest
from pyproj import CRS

from weitsicht.exceptions import CRSInputError, CRSnoZaxisError, MappingError
from weitsicht.mapping.base_class import MappingType
from weitsicht.mapping.map_trimesh import MappingTrimesh

DATA_DIR = Path(__file__).parent.resolve() / "data"


@pytest.fixture
def mesh():

    # First test wrong datafiles:
    # Test wrong path
    with pytest.raises(FileNotFoundError):
        MappingTrimesh(DATA_DIR / "wrong_decimated_160k_31256.ply")

    # Test non supported file by rasterio
    with pytest.raises(MappingError):
        MappingTrimesh(__file__)

    # Test CRS system without z Axis

    trimesh_mapper = MappingTrimesh(DATA_DIR / "decimated_160k_31256.ply", crs=CRS("EPSG:31256+5778"))

    assert int(trimesh_mapper.type)
    # CRS.proj_crs_get_coordinate_system

    return trimesh_mapper


def test_type(mesh: MappingTrimesh):
    assert mesh.type == MappingType.Trimesh


def test_from_dict():
    mesh_path = DATA_DIR / "decimated_160k_31256.ply"
    crs = CRS(31256)

    # Test "type" key missing
    with pytest.raises(KeyError):
        dict_test = {
            "type_wrong": MappingType.Trimesh.fullname,
            "mesh_filepath": mesh_path.as_posix(),
            "crs": crs.to_wkt(),
            "coordinate_shift": [0, 0, 0],
        }
        MappingTrimesh.from_dict(dict_test)

    # Test wrong "type" string
    with pytest.raises(ValueError):
        dict_test = {
            "type": "wrong",
            "mesh_filepath": mesh_path,
            "crs": crs.to_wkt(),
            "coordinate_shift": [0, 0, 0],
        }
        MappingTrimesh.from_dict(dict_test)

    # Test "mesh_filepath" key missing
    with pytest.raises(KeyError):
        dict_test = {
            "type": MappingType.Trimesh.fullname,
            "crs": crs.to_wkt(),
            "coordinate_shift": [0, 0, 0],
        }
        MappingTrimesh.from_dict(dict_test)

    # Test that CRS has no z Axis defined
    with pytest.raises(CRSnoZaxisError):
        dict_test = {
            "type": MappingType.Trimesh.fullname,
            "mesh_filepath": mesh_path,
            "crs": crs.to_wkt(),
            "coordinate_shift": [0, 0, 0],
        }
        MappingTrimesh.from_dict(dict_test)

    # Test "crs" can not be loaded
    with pytest.raises(CRSInputError):
        dict_test = {
            "type": MappingType.Trimesh.fullname,
            "mesh_filepath": mesh_path,
            "crs": "asfdasf",
            "coordinate_shift": [0, 0, 0],
        }
        MappingTrimesh.from_dict(dict_test)

    dict_test = {
        "type": MappingType.Trimesh.fullname,
        "mesh_filepath": mesh_path,
        "crs": CRS("EPSG:31256+5778").to_wkt(),
        "coordinate_shift": None,
    }

    _ = MappingTrimesh.from_dict(dict_test)


def test_param_dict(mesh: MappingTrimesh):
    mesh_path = DATA_DIR / "decimated_160k_31256.ply"
    crs = CRS("EPSG:31256+5778")

    mesh_dict = mesh.param_dict
    dict_test = {
        "type": MappingType.Trimesh.fullname,
        "mesh_filepath": mesh_path.as_posix(),
        "crs": crs.to_wkt(),
        "coordinate_shift": None,
    }

    assert dict_test == mesh_dict

    mesh.translation = np.array([0.0, 200.0, 0.0])
    mesh_dict = mesh.param_dict

    dict_test = {
        "type": MappingType.Trimesh.fullname,
        "mesh_filepath": mesh_path.as_posix(),
        "crs": crs.to_wkt(),
        "coordinate_shift": [0.0, 200.0, 0.0],
    }

    assert dict_test == mesh_dict


def test_map_coordinates_from_rays(mesh: MappingTrimesh):
    ray_start = np.array([[2643.05, 342715.09, 4.95], [2643.20, 342717.07, 4.91]])

    ray_vector = np.array([[1, 0, 0], [1, 0, 0]])

    result = mesh.map_coordinates_from_rays(ray_start_crs_s=ray_start, ray_vectors_crs_s=ray_vector, crs_s=mesh.crs)
    assert result.ok is False

    ray_vector = np.array([[-1, 0, 0], [-1, 0, 0]])
    result = mesh.map_coordinates_from_rays(ray_start_crs_s=ray_start, ray_vectors_crs_s=ray_vector, crs_s=mesh.crs)

    assert result.ok is True


def test_map_heights_from_coordinates(mesh: MappingTrimesh):

    # The second coordinate is on a hole on the mesh
    # Therefore it should have the mask False for the second coordinate
    coordinates_to_check = np.array([[2638.8261, 342718.072, 100], [2639.06, 342719.40295, 100]])

    # Hand picked that coordinates
    height_ref = np.array([3.3225, 3.37239])

    result = mesh.map_heights_from_coordinates(coordinates_crs_s=coordinates_to_check, crs_s=mesh.crs)

    assert result.ok is True
    assert result.coordinates.shape == (2, 3)
    np.testing.assert_array_equal(result.mask, np.array([True, False]))

    # This coordinates work well
    coordinates_to_check = np.array([[2638.8261, 342718.072, 100], [2639.047977, 342719.404663, 100]])
    result = mesh.map_heights_from_coordinates(coordinates_crs_s=coordinates_to_check, crs_s=mesh.crs)

    assert result.ok is True
    assert result.coordinates.shape == (2, 3)
    assert np.all(result.mask)
    if np.max(np.abs(result.coordinates[:, 2] - height_ref)) > 0.001:
        raise AssertionError()



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

# We follow the specifications from here, but there is nothing standardized at all
# https://support.pix4d.com/hc/en-us/articles/360016450032

import logging

import numpy as np
from numpy import cos, sin
from pyproj import CRS
from pyproj.crs.crs import CompoundCRS

from weitsicht.exceptions import CoordinateTransformationError
from weitsicht.metadata.metadata_results import (
    EORFromMetaResult,
    EORFromMetaResultSuccess,
    MetadataIssue,
)
from weitsicht.metadata.tag_systems.tag_base import MetaTagAll
from weitsicht.transform.rotation import Rotation
from weitsicht.transform.utm_converter import point_convert_utm_wgs84_egm2008
from weitsicht.utils import ResultFailure

logger = logging.getLogger(__name__)

aircraft_notation_to_front_notation = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
swap_ned_to_enu_coo_system = np.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]])
swap_body_cam = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
swap_body_cam_gimbal = np.array([[0, 0, -1], [-1, 0, 0], [0, 1, 0]])


def _deg(val):
    return float(val) * np.pi / 180.0


def eor_from_meta(
    tags: MetaTagAll,
    crs: CRS | None = None,
    vertical_ref: str = "ellipsoidal",
    height_rel: float = 0.0,
) -> EORFromMetaResult:
    """Extract exterior orientation (position + rotation + CRS) from metadata tags.

    The returned position is expressed in a local projected coordinate system derived from the
    input CRS (typically WGS84) by converting to WGS84-UTM with EGM2008 heights.

    If XMP tags ``HorizCS``/``VertCS`` are missing, defaults to WGS84 (EPSG:4979) with ellipsoidal heights.

    :param tags: Grouped metadata values (e.g. from ``MetaTagsBase.get_all()``).
    :type tags: MetaTagAll
    :param crs: Optional override CRS for interpreting the position tags, defaults to ``None``.
    :type crs: CRS | None
    :param vertical_ref: Vertical reference mode: ``ellipsoidal``, ``orthometric``, or ``relative``.
        ``relative`` uses ``RelativeAltitude`` + ``height_rel``.
    :type vertical_ref: str
    :param height_rel: Reference height (meters) for ``vertical_ref='relative'``, defaults to ``0.0``.
    :type height_rel: float
    :return: Successful EOR result or a failure result.
    :rtype: EORFromMetaResult
    """

    crs_source: CRS | None = crs
    position = np.array([0, 0, 0])

    lon = tags.gps.gps_longitude
    lat = tags.gps.gps_latitude
    alt = tags.gps.gps_altitude
    lon_ref = tags.gps.gps_longitude_ref or "E"
    lat_ref = tags.gps.gps_latitude_ref or "N"

    if lon is not None and lat is not None and alt is not None:
        x_exif = lon
        if lon_ref == "W":
            x_exif = -x_exif
        y_exif = lat
        if lat_ref == "S":
            y_exif = -y_exif
        z_exif = alt
    else:
        return ResultFailure(ok=False, error="Standard GPS tags are missing", issues={MetadataIssue.MISSING_GPS})

    #
    rel_z_exif = None
    if vertical_ref == "relative":
        rel_z_exif = tags.z_alternatives.rel_altitude
        if rel_z_exif is not None:
            z_exif = height_rel + float(rel_z_exif)

    # Determine CRS for interpreting the input coordinates if not overridden by the user.
    if crs_source is None:
        if vertical_ref == "orthometric":
            crs_hor_exif = 4326
            crs_vert_exif = 3855

        else:
            crs_hor_exif = tags.crs.horiz_cs or 4979
            crs_vert_exif = tags.crs.vert_cs or "ellipsoidal"

        if tags.crs.horiz_cs is not None:
            crs_hor_exif = tags.crs.horiz_cs
            crs_vert_exif = tags.crs.vert_cs or "ellipsoidal"

        # This is now an override if relative height is specified:
        # Then we will
        if rel_z_exif is not None:
            crs_hor_exif = 4326
            crs_vert_exif = 3855

        if crs_vert_exif == "ellipsoidal":
            crs_source = CRS(crs_hor_exif).to_3d()
        else:
            crs_source = CompoundCRS(
                str(crs_hor_exif) + "+" + str(crs_vert_exif),
                [crs_hor_exif, crs_vert_exif],
            )

    assert crs_source is not None
    try:
        x, y, z, crs_result = point_convert_utm_wgs84_egm2008(crs_source, x_exif, y_exif, z_exif)
    except (ValueError, CoordinateTransformationError) as err:
        return ResultFailure(
            ok=False, error=str(err), issues={MetadataIssue.TRANSFORMATION_FAILED, MetadataIssue.EOR_FAILED}
        )

    position = np.array([x, y, z])

    # Orientation

    xmp_cam_pitch = tags.orientation.xmp_camera_pitch
    xmp_cam_roll = tags.orientation.xmp_camera_roll
    xmp_cam_yaw = tags.orientation.xmp_camera_yaw

    maker_notes_cam_pitch = tags.orientation.maker_notes_camera_pitch
    maker_notes_cam_roll = tags.orientation.maker_notes_camera_roll
    maker_notes_cam_yaw = tags.orientation.maker_notes_camera_yaw

    xmp_pitch_xyz = tags.orientation.xmp_pitch
    xmp_roll_xyz = tags.orientation.xmp_roll
    xmp_yaw_xyz = tags.orientation.xmp_yaw

    maker_pitch_xyz = tags.orientation.maker_notes_pitch
    maker_roll_xyz = tags.orientation.maker_notes_roll
    maker_yaw_xyz = tags.orientation.maker_notes_yaw

    xmp_gimbal_roll = tags.orientation.xmp_gimbal_roll_deg
    xmp_gimbal_yaw = tags.orientation.xmp_gimbal_yaw_deg
    xmp_gimbal_pitch = tags.orientation.xmp_gimbal_pitch_deg

    maker_notes_gimbal_roll = tags.orientation.maker_notes_gimbal_roll_deg
    maker_notes_gimbal_yaw = tags.orientation.maker_notes_gimbal_yaw_deg
    maker_notes_gimbal_pitch = tags.orientation.maker_notes_gimbal_pitch_deg

    angle_in_direction_of_view = False
    pitch = None
    if xmp_cam_roll is not None and xmp_cam_yaw is not None and xmp_cam_pitch is not None:
        angle_in_direction_of_view = True
        roll = _deg(xmp_cam_roll)
        yaw = _deg(xmp_cam_yaw)
        pitch = _deg(xmp_cam_pitch)
    elif maker_notes_cam_roll is not None and maker_notes_cam_yaw is not None and maker_notes_cam_pitch is not None:
        angle_in_direction_of_view = True
        roll = _deg(maker_notes_cam_roll)
        yaw = _deg(maker_notes_cam_yaw)
        pitch = _deg(maker_notes_cam_pitch)
    elif xmp_roll_xyz is not None and xmp_yaw_xyz is not None and xmp_pitch_xyz is not None:
        roll = _deg(xmp_roll_xyz)
        yaw = _deg(xmp_yaw_xyz)
        pitch = _deg(xmp_pitch_xyz)
    elif maker_roll_xyz is not None and maker_yaw_xyz is not None and maker_pitch_xyz is not None:
        roll = _deg(maker_roll_xyz)
        yaw = _deg(maker_yaw_xyz)
        pitch = _deg(maker_pitch_xyz)
    elif xmp_gimbal_roll is not None and xmp_gimbal_yaw is not None and xmp_gimbal_pitch is not None:
        angle_in_direction_of_view = True
        roll = _deg(xmp_gimbal_roll)
        yaw = _deg(xmp_gimbal_yaw)
        pitch = _deg(xmp_gimbal_pitch)
    elif (
        maker_notes_gimbal_roll is not None
        and maker_notes_gimbal_yaw is not None
        and maker_notes_gimbal_pitch is not None
    ):
        angle_in_direction_of_view = True
        roll = _deg(maker_notes_gimbal_roll)
        yaw = _deg(maker_notes_gimbal_yaw)
        pitch = _deg(maker_notes_gimbal_pitch)
    else:
        return ResultFailure(
            ok=False,
            error="Orientation angles are not in meta-data",
            issues={MetadataIssue.MISSING_ORIENTATION, MetadataIssue.EOR_FAILED},
        )

    if pitch is None:
        return ResultFailure(
            ok=False,
            error="Orientation angles are not in meta-data",
            issues={MetadataIssue.MISSING_ORIENTATION, MetadataIssue.EOR_FAILED},
        )
    # Rotation of IMAGE still in Body System
    rot_sys = np.array(
        [
            [
                cos(pitch) * cos(yaw),
                sin(roll) * sin(pitch) * cos(yaw) - cos(roll) * sin(yaw),
                cos(roll) * sin(pitch) * cos(yaw) + sin(roll) * sin(yaw),
            ],
            [
                cos(pitch) * sin(yaw),
                sin(roll) * sin(pitch) * sin(yaw) + cos(roll) * cos(yaw),
                cos(roll) * sin(pitch) * sin(yaw) - sin(roll) * cos(yaw),
            ],
            [-sin(pitch), sin(roll) * cos(pitch), cos(roll) * cos(pitch)],
        ]
    )

    # Bring rotation into the cameras coordinate system. X left, Y top, Z backwards of viewing direction
    rot_enu_body = (swap_ned_to_enu_coo_system @ rot_sys) @ aircraft_notation_to_front_notation

    if angle_in_direction_of_view:
        rot_enu_cam = rot_enu_body @ swap_body_cam_gimbal
    else:
        rot_enu_cam = rot_enu_body @ swap_body_cam

    orientation = Rotation(rot_enu_cam)

    return EORFromMetaResultSuccess(ok=True, position=position, orientation=orientation, crs=crs_result)



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""OpenCV-based pinhole camera model implementation."""

from __future__ import annotations

import logging
from typing import Literal

import numpy as np

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.camera.camera_types import CameraType
from weitsicht.utils import ArrayNx2, ArrayNx3, Vector2D, to_array_nx3

__all__ = ["CameraOpenCVPerspective"]


logger = logging.getLogger(__name__)


class CameraOpenCVPerspective(CameraBasePerspective):
    """OpenCV perspective (pinhole) camera model.

    See also OpenCV calib3d documentation:
    https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html
    """

    def __init__(
        self,
        width: int,
        height: int,
        fx: float,
        fy: float,
        cx: float = 0.0,
        cy: float = 0.0,
        k1: float = 0.0,
        k2: float = 0.0,
        k3: float = 0.0,
        k4: float = 0.0,
        p1: float = 0.0,
        p2: float = 0.0,
    ):
        """Initialize an OpenCV pinhole camera model.

        :param width: Calibration image width in pixels.
        :type width: int
        :param height: Calibration image height in pixels.
        :type height: int
        :param fx: Focal length in x direction in pixels.
        :type fx: float
        :param fy: Focal length in y direction in pixels.
        :type fy: float
        :param cx: Principal point x coordinate in pixels, defaults to ``0.0`` (interpreted as ``width/2 - 0.5``).
        :type cx: float
        :param cy: Principal point y coordinate in pixels, defaults to ``0.0`` (interpreted as ``height/2 - 0.5``).
        :type cy: float
        :param k1: Radial distortion coefficient, defaults to ``0.0``.
        :type k1: float
        :param k2: Radial distortion coefficient, defaults to ``0.0``.
        :type k2: float
        :param k3: Radial distortion coefficient, defaults to ``0.0``.
        :type k3: float
        :param k4: Radial distortion coefficient, defaults to ``0.0``.
        :type k4: float
        :param p1: Tangential distortion coefficient, defaults to ``0.0``.
        :type p1: float
        :param p2: Tangential distortion coefficient, defaults to ``0.0``.
        :type p2: float
        :raises ValueError: If ``width``, ``height``, ``fx`` or ``fy`` are not positive.
        :raises TypeError: If any parameter has an incompatible type.
        """

        if width <= 0 or height <= 0 or fx <= 0 or fy <= 0:
            raise ValueError("with, height, fx and fy are not allowed to be <= 0")

        super().__init__(width=width, height=height)

        self.fx = fx
        self.fy = fy
        self.cx = cx
        self.cy = cy

        # OpenCV is defined as the origin is in the center pixel
        # Therefore center point is shifted 0.5 pixel if no principal point is given
        if self.cx == 0.0:
            self.cx = width / 2 - 0.5
        if self.cy == 0.0:
            self.cy = height / 2 - 0.5

        self.k1 = k1
        self.k2 = k2
        self.k3 = k3
        self.k4 = k4
        self.p1 = p1
        self.p2 = p2

    @classmethod
    def from_dict(cls, param_dict: dict) -> CameraOpenCVPerspective:
        """Create a camera model from a parameter dictionary.

        Required keys are:
        - ``calib_width``
        - ``calib_height``
        - ``fx``
        - ``fy``

        Optional keys are:
        - ``cx``, ``cy``
        - ``k1``, ``k2``, ``k3``, ``k4``
        - ``p1``, ``p2``

        :param param_dict: Dictionary with camera parameters.
        :type param_dict: dict
        :return: Camera model instance.
        :rtype: CameraOpenCVPerspective
        :raises KeyError: If required keys are missing.
        :raises ValueError: If values are invalid (e.g. non-positive dimensions/focal lengths).
        :raises TypeError: If values have incompatible types.
        """

        try:
            camera = cls(
                width=param_dict["calib_width"],
                height=param_dict["calib_height"],
                fx=param_dict["fx"],
                fy=param_dict["fy"],
                cx=param_dict.get("cx", 0.0),
                cy=param_dict.get("cy", 0.0),
                k1=param_dict.get("k1", 0.0),
                k2=param_dict.get("k2", 0.0),
                k3=param_dict.get("k3", 0.0),
                k4=param_dict.get("k4", 0.0),
                p1=param_dict.get("p1", 0.0),
                p2=param_dict.get("p2", 0.0),
            )

            return camera
        except (KeyError, ValueError) as err:
            raise err

    @property
    def type(self) -> Literal[CameraType.OpenCV]:
        return CameraType.OpenCV

    @property
    def param_dict(self) -> dict:

        param = {
            "type": self.type.fullname,
            "calib_width": self.calibration_width,
            "calib_height": self.calibration_height,
            "fx": self.fx,
            "fy": self.fy,
            "cx": self.cx,
            "cy": self.cy,
            "k1": self.k1,
            "k2": self.k2,
            "k3": self.k3,
            "k4": self.k4,
            "p1": self.p1,
            "p2": self.p2,
        }
        return param

    @property
    def focal_length_for_gsd_in_pixel(self) -> float:
        return float(self.fx / 2.0 + self.fy / 2.0)

    @property
    def _origin(self) -> Vector2D:
        """Return the origin offset for OpenCV pixel conventions.

        OpenCV defines the origin at the center of the top-left pixel, whereas
        this project uses the top-left pixel edge as (0, 0).

        :return: Origin offset in pixels.
        :rtype: Vector2D
        """
        return np.array([0.5, 0.5])

    @property
    def _principal_point(self) -> Vector2D:
        return np.array([self.cx, self.cy])

    def _camara_crs_to_image_crs(self, pts_camera_crs: ArrayNx3) -> ArrayNx2:
        """Project camera CRS coordinates to undistorted image pixel coordinates.

        :param pts_camera_crs: Points in camera CRS.
        :type pts_camera_crs: ArrayNx3
        :return: Undistorted image pixel coordinates.
        :rtype: ArrayNx2
        """

        _pts_camera_crs = to_array_nx3(pts_camera_crs)

        # Change to openCV CRS definition
        _pts_camera_crs[:, 1] = -_pts_camera_crs[:, 1]
        _pts_camera_crs[:, 2] = -_pts_camera_crs[:, 2]

        mat_proj = np.array([[self.fx, 0, self.cx], [0, self.fy, self.cy], [0.0, 0.0, 1.0]])

        points_image_crs = np.matmul(mat_proj, _pts_camera_crs.T).T
        points_image_crs = (points_image_crs[:, 0:2].T / points_image_crs[:, 2]).T

        return points_image_crs

    def _image_crs_to_vector_in_camera_crs(self, pts_image: ArrayNx2) -> ArrayNx3:
        """Convert undistorted image pixels to unit vectors in camera CRS.

        :param pts_image: Undistorted image pixel coordinates.
        :type pts_image: ArrayNx2
        :return: Unit direction vectors in camera CRS.
        :rtype: ArrayNx3
        """
        # vector through the pixel
        line_vec = -np.ones((pts_image.shape[0], 3))
        line_vec_xy = pts_image - np.array([self.cx, self.cy])
        line_vec_xy = line_vec_xy / np.array([self.fx, -self.fy])
        line_vec[:, 0:2] = line_vec_xy

        line_vec = np.divide(line_vec.T, np.linalg.norm(line_vec, axis=1)).T

        return line_vec

    def _to_distorted_points(self, pts_undistorted_pixel: ArrayNx2) -> ArrayNx2:
        """Calculate distorted points from undistorted.
        Polynomial functions could map extreme outside points inside due too polynom terms.
        Therefore, we should check for validity first the undistorted points if within border

        :param pts_undistorted_pixel: Undistorted image pixel coordinates.
        :type pts_undistorted_pixel: ArrayNx2
        :return: Distorted image pixel coordinates.
        :rtype: ArrayNx2
        """

        fxy = np.array([self.fx, self.fy])
        p1_p2 = np.array([self.p1, self.p2])
        principle_point = np.array([self.cx, self.cy])

        # As this function is starting from the undistorted we first to some steps to get the x',y' used in opencv
        # x' = Xc/Zc
        # y' = Yc/Zc
        # these are the pts_
        pts_ = np.divide(pts_undistorted_pixel[:, 0:2] - principle_point, fxy)
        # r² = x'² + y'²
        r_2 = np.sum(np.power(pts_, 2), axis=1)
        radial_fac = r_2 * self.k1 + (r_2**2) * self.k2 + (r_2**3) * self.k3 + (r_2**4) * self.k4
        radial = ((pts_ * fxy).T * radial_fac).T

        # tangential x = 2 p1 x' y' + p2 (r2 + 2 x'²)
        # tangential y = 2 p2 x' y' + p1 (r2 + 2 y'²)
        tangential = (
            2 * fxy * pts_ * np.flip(pts_, axis=1) * p1_p2 + np.flip(p1_p2) * fxy * (r_2 + 2 * np.power(pts_, 2).T).T
        )

        # x'' = x' *
        # u = fx * x'' + cx
        pts_distorted = pts_undistorted_pixel + radial + tangential

        return pts_distorted

    def _to_undistorted_points(self, pts_distorted_pixel: ArrayNx2) -> ArrayNx2:
        """Calculate undistorted points from distorted image pixel coordinates.

        :param pts_distorted_pixel: Distorted image pixel coordinates.
        :type pts_distorted_pixel: ArrayNx2
        :return: Undistorted image pixel coordinates.
        :rtype: ArrayNx2
        """

        # TODO There might be a problem with extremely high distortion values. Then the iteration by using the other
        #  direction as approx values will not work. Once I saw that than max_diff or pts_undistorted becomes np.nan

        # As the inversion is non-linear and can not be inverted one option is to use iteration
        # using the _to_distorted_point function to minimize the differences
        pts_undistorted = pts_distorted_pixel + (pts_distorted_pixel - self._to_distorted_points(pts_distorted_pixel))
        max_diff = np.max(np.linalg.norm(pts_distorted_pixel - self._to_distorted_points(pts_undistorted), axis=1))
        iteration = 0
        while max_diff > 0.001 and iteration < 30:
            iteration += 1
            pts_undistorted = pts_undistorted + (pts_distorted_pixel - self._to_distorted_points(pts_undistorted))
            max_diff = np.max(
                np.linalg.norm(
                    pts_distorted_pixel - self._to_distorted_points(pts_undistorted),
                    axis=1,
                )
            )

        return pts_undistorted



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

import numpy as np
import pytest

from weitsicht.camera.opencv_perspective import CameraOpenCVPerspective


@pytest.fixture
def camera_class():
    cam_class = CameraOpenCVPerspective(
        width=5472,
        height=3648,
        fx=3753.0,
        fy=3753.0,
        cx=2721.2,
        cy=1813.7,
        k1=-0.0082,
        k2=0.00053,
        k3=0.0089,
        p1=-0.0015,
        p2=-0.00086,
    )
    assert int(cam_class.type)
    return cam_class


def test_generate_distortion_border(camera_class):
    """

    OpenCV is used as reference:
    import cv2
    import numpy as np
    img_points = np.array([[0., 0.], [912., 0.], [1824., 0.],
                       [2736., 0.], [3648., 0.], [4560., 0.],
                       [5472., 0.], [5472., 608.], [5472., 1216.],
                       [5472., 1824.], [5472., 2432.], [5472., 3040.],
                       [5472., 3648.], [4560., 3648.], [3648., 3648.],
                       [2736., 3648.], [1824., 3648.], [912., 3648.],
                       [0., 3648.], [0., 3040.], [0., 2432.],
                       [0., 1824.], [0., 1216.], [0., 608.]])
    # We need to shift as we generate our border at the pixel edge and opencv is defined at pixel center point
    img_points -= 0.5
    cm = np.array([[3753,0,2721.2],
                    [0,3753, 1813.7],
                    [0,0,1]])
    distCoeff = np.array([-0.0082, 0.00053, -0.0015, -0.00086, 0.0089])
    pts = cv2.undistortPoints(img_points, cm, distCoeff, P= cm)
    print(pts[:,0])

    """

    ref_unidostorted = np.array(
        [
            [3.6950856651737922, 4.926089041198338],
            [912.0498619053583, 1.166468531321243],
            [1824.2054469344891, 0.6661106176857174],
            [2736.258629429585, 0.21640851877396017],
            [3649.490791714201, -0.8701864992663104],
            [4565.120716488502, -1.8470306251776947],
            [5478.748812925189, 0.7278971900686884],
            [5481.816368398053, 607.5286790909195],
            [5483.398896310726, 1216.5151259677384],
            [5484.714312680481, 1826.6059687681761],
            [5486.112210815349, 2437.5928060337123],
            [5487.1874942863315, 3049.236019587731],
            [5486.611337234232, 3660.3181640614475],
            [4570.6074982813325, 3659.7546166435654],
            [3652.2821347336135, 3656.736160592074],
            [2736.32678114928, 3654.926858533194],
            [1821.5640557720176, 3655.111288891173],
            [906.7493917432471, 3656.5548025336248],
            [-3.949116069520187, 3655.8320076016635],
            [-5.12373219527899, 3046.095130475993],
            [-4.472306534030395, 2435.939819493362],
            [-3.2485788179346855, 1826.4876245091502],
            [-1.834510860566752, 1217.9239326802656],
            [0.09759060037094969, 610.405332026314],
        ]
    )

    # The ref_unidostorted is generate with opencv so we shift by 0.5
    ref_unidostorted += 0.5

    undistorted_border = camera_class._generate_distortion_border(points_between=5)
    print("Dist max[pix]:", np.max(undistorted_border - ref_unidostorted))
    assert np.allclose(undistorted_border, ref_unidostorted, atol=1e-3, rtol=0)


def test_pixel_image_to_camera_crs(camera_class):
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    prc_pt = camera_class.principal_point(img_size)

    # The principal point will not be distorted
    result = camera_class.pixel_image_to_camera_crs(prc_pt, img_size)
    assert result[0].tolist() == [0, 0, -1]


def test_pts_camara_crs_to_image_pixel(camera_class):
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    prc_pt = camera_class.principal_point(img_size)

    # The principal point will not be distorted
    result = camera_class.pts_camara_crs_to_image_pixel(np.array([0, 0, 1]), img_size)

    assert result[0].tolist() == prc_pt.tolist()


def test_distorted_to_undistorted(camera_class):
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    prc_pt = camera_class.principal_point(img_size)
    result = camera_class.distorted_to_undistorted(prc_pt, img_size)
    assert result[0].tolist() == prc_pt.tolist()
    test_points = np.array([[100, 100]])
    result = camera_class.distorted_to_undistorted(test_points, img_size)
    result_2 = camera_class.undistorted_to_distorted(result, img_size)
    assert np.linalg.norm(test_points - result_2) < 0.1


def test_undistorted_to_distorted(camera_class):
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    prc_pt = camera_class.principal_point(img_size)

    # The principal point will not be distorted
    result = camera_class.distorted_to_undistorted(prc_pt, img_size)

    assert result[0].tolist() == prc_pt.tolist()


def test_image_points_inside(camera_class):
    # Test with image points in camera calibration size
    image_points = np.array(
        [
            [
                0,
                0,
            ],
            [100, 100],
        ]
    )
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    valid_index = camera_class.undistorted_image_points_inside(image_points, image_size=img_size)

    assert not bool(valid_index[0])
    assert bool(valid_index[1])

    # Test with image size
    # these points are from opencv
    image_points = np.array([[2736.32678114928, 3654.926858533194], [2736.32678114928, 3654.926858533194]])
    image_points += 0.5

    # shift a little to be sure one is inside and one outside.
    # Undisotort iteration is about 1e-2 accurate min in the current configuration
    image_points = image_points + np.array([[0.05, 0.05], [-0.05, -0.05]])

    image_points = image_points / 2.0
    # Simulate image which is 50% of original image used for calibration
    img_size = (
        camera_class.calibration_width / 2.0,
        camera_class.calibration_height / 2.0,
    )
    valid_index = camera_class.undistorted_image_points_inside(image_points, image_size=img_size)

    assert not bool(valid_index[0])
    assert bool(valid_index[1])

    image_points = np.array(
        [
            [3.6950856651737922, 4.926089041198338],
            [3.6950856651737922, 4.926089041198338],
        ]
    )
    # The values are from opencv so we shift 0.5
    image_points += 0.5

    # We shift small pixel values. One is outside the other inside
    image_points = image_points + np.array([[0.01, 0.01], [-0.1, 0.01]])
    img_size = (camera_class.calibration_width, camera_class.calibration_height)
    valid_index = camera_class.undistorted_image_points_inside(image_points, image_size=img_size)

    assert bool(valid_index[0])
    assert not bool(valid_index[1])



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------

"""Base API for camera perspective models.

Defines :class:`~weitsicht.camera.base_perspective.CameraBasePerspective`, which
converts between image pixel coordinates and camera coordinate system vectors,
including distortion/undistortion handling.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import cast

import numpy as np
from shapely import geometry

from weitsicht.camera.camera_types import CameraType
from weitsicht.utils import (
    ArrayNx2,
    ArrayNx3,
    MaskN_,
    Vector2D,
    to_array_nx2,
    to_array_nx3,
)

__all__ = ["CameraBasePerspective"]


class PostMetaCaller(type):
    def __call__(cls, *args, **kwargs):
        # here is "before __new__ is called"
        instance = type.__call__(cls, *args, **kwargs)
        instance.__post_init__()

        return instance


class CameraBasePerspective(metaclass=PostMetaCaller):
    """Base class for perspective camera models.

    This class converts between image pixel coordinates ("image CRS") and rays in
    camera coordinates ("camera CRS"). It does not store geo-referencing; world
    coordinate transforms are handled by the image layer.

    Coordinate system conventions:
    - Image pixel CRS: x points right, y points down (origin at top-left pixel edge).
    - Camera CRS: x points right, y points up, z points backwards.

    Implementations typically apply origin shifts and scale pixel coordinates to
    a calibration size so that distortion parameters remain valid for resampled
    images.

    Subclasses must implement:
    - `from_dict`
    - `type`
    - `param_dict`
    - `focal_length_for_gsd_in_pixel`
    - `_origin`
    - `_principal_point`
    - `_to_distorted_points`
    - `_to_undistorted_points`
    - `_image_crs_to_vector_in_camera_crs`
    - `_camara_crs_to_image_crs`
    """

    def __init__(self, width: int, height: int, pts_distortion: int = 5):
        """Initialize a camera perspective model base.

        :param width: Calibration image width in pixels.
        :type width: int
        :param height: Calibration image height in pixels.
        :type height: int
        :param pts_distortion: Number of points used to generate the distortion border, defaults to ``5``.
        :type pts_distortion: int
        """

        self.calibration_width = width
        self.calibration_height = height

        # initial border
        self.distortion_border = geometry.Polygon(
            [
                [0, 0],
                [self.calibration_width, 0],
                [self.calibration_width, self.calibration_height],
                [0, self.calibration_height],
            ]
        )
        self.pts_distortion = pts_distortion

    def __post_init__(self):
        """Finalize initialization.

        Called automatically after initialization (including subclasses) to
        compute the polygon that bounds valid undistorted points.
        """
        distortion_border = self._generate_distortion_border(self.pts_distortion)
        self.distortion_border = geometry.Polygon(distortion_border)

    @classmethod
    @abstractmethod
    def from_dict(cls, param_dict: dict) -> CameraBasePerspective:
        """Create a camera model from a parameter dictionary.

        :param param_dict: Dictionary as produced by :attr:`param_dict`.
        :type param_dict: dict
        :return: Camera model instance.
        :rtype: CameraBasePerspective
        """
        # Has to be implemented for each camera Class
        pass

    @property
    @abstractmethod
    def param_dict(self) -> dict:
        """Return camera parameters as a dictionary.

        The returned dictionary must be compatible with :meth:`from_dict`.

        :return: Camera parameters.
        :rtype: dict
        """
        # Has to be implemented for each camera Class
        pass

    @property
    @abstractmethod
    def type(self) -> CameraType:
        """Return the camera model type.

        :return: Camera model type.
        :rtype: CameraType
        """
        # Has to be implemented for each camera Class
        ...

    @property
    @abstractmethod
    def focal_length_for_gsd_in_pixel(self) -> float:
        """Returns the focal length in pixel to be used for approximate gsd calculation

        :return: Focal length in pixel.
        :rtype: float
        """
        # Has to be implemented for each camera Class
        pass

    @property
    @abstractmethod
    def _origin(self) -> Vector2D:
        """One has to implement the pixel origin of the camera model as offset to the used
        pixel system. Our system is using the upper left corner as 0,0
        For example OpenCVs pixel coordinate system is defined with left upper CENTER is (0,0) therefore
        we need to apply an offset and return np.array([0.5,0.5])

        :return: Origin offset in pixels.
        :rtype: Vector2D
        """
        pass

    @property
    @abstractmethod
    def _principal_point(self) -> Vector2D:
        """principal point in camera size

        :return: Principal point in camera calibration size.
        :rtype: Vector2D
        """
        # Has to be implemented for each camera Class
        pass

    def _generate_distortion_border(self, points_between: int = 5) -> ArrayNx2:
        """Returns the border of the undistorted image, which is used as border for points which
        are valid to calculate the distorted ones (The distorted border is the rectangular image itself)

        :param points_between: Number of points between corner points, defaults to ``5``.
        :type points_between: int
        :return: ArrayNx2 with the undistorted pixel of the border points
        :rtype: ArrayNx2
        """

        img_points = np.vstack(
            (
                np.linspace(
                    [0, 0],
                    [self.calibration_width, 0],
                    points_between + 1,
                    axis=0,
                    endpoint=False,
                ),
                np.linspace(
                    [self.calibration_width, 0],
                    [self.calibration_width, self.calibration_height],
                    points_between + 1,
                    axis=0,
                    endpoint=False,
                ),
                np.linspace(
                    [self.calibration_width, self.calibration_height],
                    [0, self.calibration_height],
                    points_between + 1,
                    axis=0,
                    endpoint=False,
                ),
                np.linspace(
                    [0, self.calibration_height],
                    [0, 0],
                    points_between + 1,
                    axis=0,
                    endpoint=False,
                ),
            )
        )

        img_points = cast(ArrayNx2, img_points)
        distortion_border = self.distorted_to_undistorted(img_points, (self.calibration_width, self.calibration_height))

        return distortion_border

    @abstractmethod
    def _to_distorted_points(self, pts_undistorted_pixel: ArrayNx2) -> ArrayNx2:
        """function for calculating distorted points from undistorted ones in original calibration image size

        :param pts_undistorted_pixel: Array of (nx2) with the image coordinates
        :type pts_undistorted_pixel: ArrayNx2
        :return: Array with distorted points
        :rtype: ArrayNx2
        """
        # Has to be implemented for each camera Class
        pass

    @abstractmethod
    def _to_undistorted_points(self, pts_distorted_pixel: ArrayNx2) -> ArrayNx2:
        """function for calculating undistorted points from distorted ones

        :param pts_distorted_pixel: Array of (nx2) with the image coordinates
        :type pts_distorted_pixel: ArrayNx2
        :return: ArrayNx2 with undistorted points
        :rtype: ArrayNx2
        """
        # Has to be implemented for each camera Class
        pass

    @abstractmethod
    def _image_crs_to_vector_in_camera_crs(self, pts_image: ArrayNx2) -> ArrayNx3:
        """function for calculating vector in camera crs of undistorted image points

        :param pts_image: Array of (nx2) with the image coordinates
        :type pts_image: ArrayNx2
        :return: Array of (Nx3) with vectors of the line of sight of image positions
        :rtype: ArrayNx3
        """
        # Has to be implemented for each camera Class
        pass

    @abstractmethod
    def _camara_crs_to_image_crs(self, pts_camera_crs: ArrayNx3) -> ArrayNx2:
        """function for calculating undistorted points from points in camera crs

        :param pts_camera_crs: Array of (nx3) with the camera CRS coordinates
        :type pts_camera_crs: ArrayNx3
        :return: Array of size (N,2) with undistorted points
        :rtype: ArrayNx2
        """
        # Has to be implemented for each camera Class
        pass

    @property
    def origin(self) -> Vector2D:
        return self._origin

    def principal_point(self, image_size: tuple[int, int]) -> Vector2D:
        """Return the principal point in the given image size.

        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :return: Principal point in image pixel CRS.
        :rtype: Vector2D
        """
        return self._pixel_calibration_to_image_size(self._principal_point, image_size)[0]

    def _pixel_image_to_calibration_size(self, pixel_from_image: ArrayNx2, image_size: tuple[int, int]) -> ArrayNx2:
        """Calculate scaled image coordinates to calibration size

        :param pixel_from_image: Array of (nx2) with the image coordinates
        :type pixel_from_image: ArrayNx2
        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :return: Array with the image positions scaled to calibration image size
        :rtype: ArrayNx2
        """
        _image_size = np.array(image_size)
        _pixel_from_image = to_array_nx2(pixel_from_image)

        px_normalized = _pixel_from_image / _image_size
        px_calibration_size = px_normalized * np.array([self.calibration_width, self.calibration_height])
        px_calibration_size -= self._origin
        return px_calibration_size

    def _pixel_calibration_to_image_size(self, pixel_from_calib_size: ArrayNx2, image_size: tuple[int, int]):
        """Calculate scaled image coordinates to image file size

        :param pixel_from_calib_size: Array of (nx2) with the image coordinates
        :type pixel_from_calib_size: ArrayNx2
        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :return: Array with the image positions scaled to file image size
        :rtype: ArrayNx2
        """
        _image_size = np.asarray(image_size)

        pixel_from_calib_size = to_array_nx2(pixel_from_calib_size)

        img_calib_size = np.array([self.calibration_width, self.calibration_height])
        px_normalized = pixel_from_calib_size / img_calib_size
        px_image_size = px_normalized * _image_size

        # Correct px image size for origin shift from the camera model
        px_image_size += self.origin
        return px_image_size

    def undistorted_image_points_inside(self, points_image_crs: ArrayNx2, image_size: tuple[int, int]) -> MaskN_:
        """Return a mask indicating whether undistorted image points are valid.

        This takes the distortion border of the camera model into account.

        :param points_image_crs: Array of (nx2) with image pixel coordinates.
        :type points_image_crs: ArrayNx2
        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :return: Boolean mask of shape (N,).
        :rtype: ``MaskN_``
        """

        pixel_calibration_size = self._pixel_image_to_calibration_size(points_image_crs, image_size=image_size)

        valid_mask = np.full((pixel_calibration_size.shape[0]), False)
        for idx, pt in enumerate(pixel_calibration_size):
            valid_mask[idx] = self.distortion_border.contains(geometry.Point(pt + self.origin))

        return valid_mask

    def pixel_image_to_camera_crs(
        self,
        points_image_crs: ArrayNx2,
        image_size: tuple[int, int],
        is_undistorted: bool = False,
    ) -> ArrayNx3:
        """Calculating vectors in camera crs of image points.
        Image points are internally scaled to calibration image size for validity of distortion parameters

        :param points_image_crs: Array of (nx2) with the image coordinates
        :type points_image_crs: ArrayNx2
        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :param is_undistorted: Whether input image coordinates are already undistorted, defaults to ``False``.
        :type is_undistorted: bool
        :return: Array of size (N,3) with vectors of the line of sight of image positions
        :rtype: ArrayNx3
        """
        points_image_crs = to_array_nx2(points_image_crs)

        points_image_scaled_to_calibration = self._pixel_image_to_calibration_size(points_image_crs, image_size)

        # Points will be undistorted.
        # points_image are in our pixel definition
        if not is_undistorted:
            points_image_scaled_to_calibration = self._to_undistorted_points(points_image_scaled_to_calibration)

        # Points will be scaled to calibration size as the camera class has its own pixel system definition
        pts_camera_crs = self._image_crs_to_vector_in_camera_crs(points_image_scaled_to_calibration)
        return pts_camera_crs

    def pts_camara_crs_to_image_pixel(
        self, points_camera_crs: ArrayNx3, image_size: tuple[int, int], to_distorted: bool = True
    ) -> ArrayNx2:
        """Calculating image pixel coordinates of points in camera crs.
        Image points are internally scaled to calibration image size for validity of distortion parameters

        :param points_camera_crs: Array of (nx3) with the camera crs coordinates
        :type points_camera_crs: ArrayNx3
        :param image_size: Size of the current image (width, height).
        :type image_size: tuple[int, int]
        :param to_distorted: Whether to return distorted image coordinates, defaults to ``True``.
        :type to_distorted: bool
        :return: Array of size (N,2) with pixels in image coordinates. Undistorted or distorted according to_distorted
                 and the camera model used.
        :rtype: ArrayNx2
        """
        points_camera_crs = to_array_nx3(points_camera_crs)

        points_image_in_camera_calib_size = self._camara_crs_to_image_crs(points_camera_crs)

        if to_distorted:
            # Here we are still in the system defined by the child camera
            points_image_in_camera_calib_size = self._to_distorted_points(points_image_in_camera_calib_size)

        points_image_scaled_to_image = self._pixel_calibration_to_image_size(
            points_image_in_camera_calib_size, image_size
        )

        return points_image_scaled_to_image

    def distorted_to_undistorted(self, points_distorted: ArrayNx2, image_size: tuple[int, int]) -> ArrayNx2:
        """Calculating undistorted image coordinates from distorted ones
        Image points are internally scaled to calibration image size for validity of distortion parameters

        :param points_distorted: Pixel coordinates in pixel CRS
        :type points_distorted: ArrayNx2
        :param image_size: Size of the image pts are from (width, height).
        :type image_size: tuple[int, int]
        :return: Array of size (Nx2) with undistorted image coordinates
        :rtype: ArrayNx2
        """
        scaled_pixel = self._pixel_image_to_calibration_size(points_distorted, image_size)
        undistorted_scaled = self._to_undistorted_points(scaled_pixel)
        undistorted_back_scaled_pixel = self._pixel_calibration_to_image_size(undistorted_scaled, image_size)
        return undistorted_back_scaled_pixel

    def undistorted_to_distorted(self, pts_undistorted: ArrayNx2, image_size: tuple[int, int]) -> ArrayNx2:
        """Calculating distorted image coordinates from undistorted ones
        Image points are internally scaled to calibration image size for validity of distortion parameters

        This function does not prove if the undistorted are inside the border of valid pixels

        :param pts_undistorted: Pixel coordinates in pixel CRS
        :type pts_undistorted: ArrayNx2
        :param image_size: Size of the image pts are from (width, height).
        :type image_size: tuple[int, int]
        :return: Array of size (Nx2) with distorted image coordinates
        :rtype: ArrayNx2
        """
        scaled_pixel = self._pixel_image_to_calibration_size(pts_undistorted, image_size)
        distorted_scaled = self._to_distorted_points(scaled_pixel)
        distorted_back_scaled_pixel = self._pixel_calibration_to_image_size(distorted_scaled, image_size)

        return distorted_back_scaled_pixel



# -----------------------------------------------------------------------
# Copyright 2026 Martin Wieser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------


"""Result objects and issues for metadata extraction and image building.

This mirrors the pattern used for projection and mapping results, but keeps metadata issues
separate from the generic :class:`weitsicht.utils.Issue` enum.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Literal

import numpy as np
from pyproj import CRS
from pyproj.crs.crs import CompoundCRS

from weitsicht.camera.base_perspective import CameraBasePerspective
from weitsicht.image.perspective import ImagePerspective
from weitsicht.transform.rotation import Rotation
from weitsicht.utils import ResultFailure

__all__ = [
    "MetadataIssue",
    "IORFromMetaResultSuccess",
    "IORFromMetaResult",
    "EORFromMetaResultSuccess",
    "EORFromMetaResult",
    "ImageFromMetaResultSuccess",
    "ImageFromMetaResult",
]


class MetadataIssue(Enum):
    """Issue codes for metadata extraction and estimation."""

    IOR_FAILED = "Metadata IOR extraction failed"
    EOR_FAILED = "Metadata EOR extraction failed"
    MISSING_GPS = "Metadata GPS tags missing"
    MISSING_ORIENTATION = "Metadata orientation tags missing"
    TRANSFORMATION_FAILED = "Metadata coordinate transformation failed"
    UNKNOWN = "unknown"


@dataclass
class IORFromMetaResultSuccess:
    """Successful IOR (intrinsics) estimation from metadata."""

    ok: Literal[True]
    camera: CameraBasePerspective
    width: int
    height: int


IORFromMetaResult = IORFromMetaResultSuccess | ResultFailure[MetadataIssue]


@dataclass
class EORFromMetaResultSuccess:
    """Successful EOR (pose) extraction from metadata."""

    ok: Literal[True]
    position: np.ndarray
    orientation: Rotation
    crs: CRS | CompoundCRS | None


EORFromMetaResult = EORFromMetaResultSuccess | ResultFailure[MetadataIssue]


@dataclass
class ImageFromMetaResultSuccess:
    """Successful image build from metadata."""

    ok: Literal[True]
    image: ImagePerspective
    ior: IORFromMetaResultSuccess
    eor: EORFromMetaResultSuccess


ImageFromMetaResult = ImageFromMetaResultSuccess | ResultFailure[MetadataIssue]
